#include "xsnmp.h"

#define Debug	(Flags & (PROG_DEBUG_HOST | PROG_DEBUG))

HOSTENTRY	*Hostlist=(HOSTENTRY *)NULL;

/*
 * Host routines
 */
struct hostent *lookuphost(hostname) char *hostname; {
    static struct hostent	hent;
    static char			address[sizeof(u_long)];
    static char			*addresses[]={address, (char *)NULL};
    u_long			hostaddr;

    if(isdigit(*hostname)) {
	/* construct our own hostent */
	hent.h_name = hostname;
	hent.h_aliases = (char **)NULL;
 	hent.h_addrtype = AF_INET;
	hent.h_length = sizeof(struct in_addr);	/* 4 bytes */
	hostaddr = inet_addr(hostname);
	bcopy(&hostaddr, address, sizeof(u_long));
	hent.h_addr_list = addresses;
	return(&hent);
    } else
	return(gethostbyname(hostname));
}

parsehost(line, ahostname, ahostmask, apingtime, acommunity)
char		*line;
char		*ahostname;
u_long		*ahostmask;
int		*apingtime;
char		*acommunity; {
    char	*cptr, *storeptr;
    char	buffer[80];
    
    cptr = skipspaces(line);
    if(! *cptr) return(0);	/* blank line */

    *ahostname = '\0';
    *ahostmask = DEFAULTNETMASK;
    *apingtime = DEFAULTPINGTIME;
    strcpy(acommunity, DEFAULTCOMMUNITY);

    sscanf(cptr, "%[^:]:%x:%d:%s", ahostname, ahostmask, apingtime
	   , acommunity);

    if(*ahostmask == 0) *ahostmask = DEFAULTNETMASK;
    if(*apingtime == 0) *apingtime = DEFAULTPINGTIME;

    return(1);
}

/*
 * Make a host style widget
 */
Widget makehostwidget(hostname)
char		*hostname; {
    register int	i;
    register Widget	w;

    i = 0;
    XtSetArg(Args[i], XtNtop, XtChainTop); i++;
    XtSetArg(Args[i], XtNbottom, XtChainTop); i++;
    XtSetArg(Args[i], XtNleft, XtChainLeft); i++;
    XtSetArg(Args[i], XtNright, XtChainLeft); i++;
    if(Flags & PROG_REVERSE)
	XtSetArg(Args[i], XtNborder, WhitePixelOfScreen(Myscrptr)), i++;
    else
	XtSetArg(Args[i], XtNborder, BlackPixelOfScreen(Myscrptr)), i++;
    XtSetArg(Args[i], XtNbackground, BlackPixelOfScreen(Myscrptr)); i++;
    XtSetArg(Args[i], XtNforeground, WhitePixelOfScreen(Myscrptr)); i++;
    if((w = XtCreateManagedWidget(hostname, labelWidgetClass, Form
				   , Args, i)) == (Widget)NULL) {
	fprintf(stderr, "Error creating widget for host '%s'\n", hostname);
	return((Widget)NULL);
    }

    return(w);
}

HOSTENTRY *newhostentry(hostname, display, hostmask, community, pinginterval
			, widget)
char	*hostname;
char	*display;
u_long	hostmask;
char	*community;
int	pinginterval;
Widget	widget; {
    register HOSTENTRY	*hptr;
    register IFENTRY	*ifptr;
    char		*hname;
    struct in_addr	hostaddr;
    struct hostent	*hent;
    char		**aptr;
    char		buffer[200];

    if((hent = lookuphost(hostname)) == (struct hostent *)NULL) {
	dispnotice(hostname, "Error looking up host!");
	return((HOSTENTRY *)NULL);
    }
    
    /* See if we already have this host */
    for(hptr = Hostlist ; hptr ; hptr = hptr->hl_next)
	for(ifptr = hptr->hl_if ; ifptr ; ifptr = ifptr->if_next)
	    for(aptr = hent->h_addr_list ; *aptr ; aptr++)
		if(ifptr->if_addr.s_addr == **(u_long **)aptr) {
		    hostaddr.s_addr = **(u_long **)aptr;
		    sprintf(buffer
			    , "New host %s has interface %s which already exists on host %s already on the display"
			    , hostname, inet_ntoa(hostaddr)
			    , hptr->hl_displayname);
		    dispnotice("Error adding host to display", buffer);
		    return((HOSTENTRY *)NULL);
		}

    if((hptr = (HOSTENTRY *)malloc(sizeof(HOSTENTRY)))
       == (HOSTENTRY *)NULL) {
	fprintf(stderr, "Can't alloc %d bytes of memory! (newhostentry)\n"
		, sizeof(HOSTENTRY));
	return((HOSTENTRY *)NULL);
    }

    hname = hent->h_name;
    hptr->hl_name = cpystr(hname);
    hptr->hl_displayname = cpystr(display);
    hptr->hl_if = (IFENTRY *)NULL;
    hptr->hl_interval = pinginterval;
    hptr->hl_lastpong = 0;
    hptr->hl_flags = FL_DOWN;
    hptr->hl_x = 0; hptr->hl_y = 0;
    hptr->hl_width = 0; hptr->hl_height = 0;
    hptr->hl_widget = widget;
    hptr->hl_community = cpystr(community);
    hptr->hl_lastsnmpif = 0;
    hptr->hl_lastsnmp = 0;
    hptr->hl_next = Hostlist;
    Hostlist = hptr;

    for(aptr = hent->h_addr_list ; *aptr ; aptr++)
	addif(hptr, **(u_long **)aptr, hostmask);

    return(hptr);
}

printhostlist() {
    register HOSTENTRY	*hptr;

    for(hptr = Hostlist ; hptr ; hptr = hptr->hl_next)
	printf("hptr=0x%x, name=%s\n", hptr, hptr->hl_name);
}

/*
 * Add a host to the Hostlist
 */
HOSTENTRY *addhost(hostname, hostmask, community, pinginterval, display)
char	*hostname;
u_long	hostmask;
char	*community;
int	pinginterval;
char	*display; {
    register Widget	widget;
    register HOSTENTRY	*hptr;
    char		*hname;

    if(display == (char *)NULL || *display == '\0') display = hostname;

    if((widget = makehostwidget(display)) == (Widget)NULL)
	return((HOSTENTRY *)NULL);

    if((hptr = newhostentry(hostname, display, hostmask, community
			    , pinginterval, widget)) == (HOSTENTRY *)NULL) {
	XtDestroyWidget(widget);
	return((HOSTENTRY *)NULL);
    }

    return(hptr);
}

freehost(hptr) HOSTENTRY *hptr; {

    if(hptr->hl_name) free(hptr->hl_name);
    if(hptr->hl_displayname) free(hptr->hl_displayname);
    if(hptr->hl_community) free(hptr->hl_community);

    XtDestroyWidget(hptr->hl_widget);

    removeifs(hptr);

    free(hptr);

    Flags |= PROG_KICKNET;	/* update net map */

    return;
}

addifs(hptr, hostname, hostmask)
HOSTENTRY	*hptr;
char		*hostname;
u_long		hostmask; {
    struct hostent	*hent;
    char		**aptr;
    struct in_addr	hostaddr;
    
    if((hent = lookuphost(hostname)) == (struct hostent *)NULL) {
	fprintf(stderr, "Error looking up hostname '%s' - skipping\n"
		, hostname);
	return;
    }
    
    for(aptr = hent->h_addr_list ; *aptr ; aptr++) {
	if(Debug) {
	    hostaddr.s_addr = *(u_long *)hent->h_addr_list[0];
	    printf("adding 2nd interface to %s as %s"
		   , hptr->hl_name
		   , inet_ntoa(hostaddr));
	    printf(" mask=%x\n", hostmask);
	}
	addif(hptr, **(u_long **)aptr, hostmask);
    }
}

removehost(hostptr) HOSTENTRY *hostptr; {
    register HOSTENTRY	*hptr;
    
    if(hostptr == Hostlist)
	Hostlist = hostptr->hl_next;
    else {
	/* Remove this host from the Hostlist */
	for(hptr = Hostlist
	    ; hptr->hl_next && hptr->hl_next != hostptr
	    ; hptr = hptr->hl_next) ;
	
	/* Found it ? */
	if(hptr->hl_next == hostptr)
	    hptr->hl_next = hptr->hl_next->hl_next;
    }
    
    freehost(hostptr);
    
    return;
}

/*
 * Read host table from xsnmp_hosts file
 */
Widget readhosts(widget) Widget widget; {
    register char	*cp, *cptr;
    char		buffer[80];
    struct in_addr	hostaddr;
    u_long		hostmask;
    char		hostname[80];
    int			pingtime, i;
    FILE		*file;
    char		translations[80];
    char		*filename;
    struct passwd	*pwent;
    char		*hname;
    char		community[40];
    HOSTENTRY		*hptr, *newhptr;

    /* First find their home directory */
    if((pwent = getpwuid(getuid())) == (struct passwd *)NULL
       || (filename = (char *)malloc(strlen(pwent->pw_dir)
				      + sizeof(HOSTLIST) + 1))
       == (char *)NULL) {
	fprintf(stderr, "Error making name of ~/%s file\n", HOSTLIST);
	exit(1);
    }
    sprintf(filename, "%s/%s", pwent->pw_dir, HOSTLIST);

    if((file = fopen(filename, "r")) == (FILE *)NULL) {
	fprintf(stderr, "Can't open hostlist file %s\n", filename);
	perror(Myname);
	return(widget);
    }
    
    while(fgets(buffer, sizeof(buffer), file)) {
	/* remove newline */
	if(Debug)
	    printf("line = '%s'\n", buffer);
	for(cp = buffer ; *cp && *cp != '\n' ; cp++) ;
	*cp = '\0';
	
	cptr = skipspaces(buffer);

	if(*cptr == '+') {	/* additional interface */
	    switch(sscanf(++cptr, "%[0-9.]", hostname)) {
	    case 0:		/* blank line? */
		continue;
	    case 1:		/* Got the additional hostname! */
		addifs(hptr, hostname, hostmask);
		break;
	    }
	    continue;		/* next input line */
	}

	if(parsehost(cptr, hostname, &hostmask, &pingtime, community) == 0)
	    continue;

	if((widget = makehostwidget(hostname))
	   == (Widget)NULL)
	    continue;

	if((newhptr = newhostentry(hostname, hostname, hostmask, community
				   , pingtime, widget)) != (HOSTENTRY *)NULL) {
	    hptr = newhptr;
	    if(Debug)
		printf("%s being setup with %d interfaces\n"
		       , hptr->hl_name, countif(hptr));
	} else
	    XtDestroyWidget(widget);
    }

    fclose(file);

    return(widget);
}

/*
 * Find a HOSTENTRY given the name
 */
HOSTENTRY *nametohost(name) char *name; {
    register HOSTENTRY	*hptr;
    char		*hname;
    struct hostent	*hent;


    if((hent = lookuphost(name)) == (struct hostent *)NULL)
	return((HOSTENTRY *)NULL);

    for(hptr = Hostlist ; hptr ; hptr = hptr->hl_next)
	if(sstrcmp(hptr->hl_name, hent->h_name) == 0)
	    return(hptr);

    return((HOSTENTRY *)NULL);
}

/*
 * Find a hostentry given the widget
 */
HOSTENTRY *widgettohost(w) Widget w; {
    register HOSTENTRY	*hptr;

    for(hptr = Hostlist ; hptr ; hptr = hptr->hl_next)
	if(hptr->hl_widget == w) return(hptr);
 
    return((HOSTENTRY *)NULL);
}


/*
 * See if there's a locations file and read it.
 */
getlocations() {
    register FILE	*file;
    register HOSTENTRY	*hostptr;
    char		buffer[80];
    char		*cptr;
    char		hostname[80];
    int			x, y;
    char		*filename;
    struct passwd	*pwent;

    /* First find their home directory */
    pwent = getpwuid(getuid());
    filename = (char *)malloc(strlen(pwent->pw_dir) + 1 + sizeof(LOCATIONS));
    if(pwent == (struct passwd *)NULL
       || filename == (char *)NULL) {
	fprintf(stderr, "Error making name of ~/%s file\n", LOCATIONS);
	return;
    }
    sprintf(filename, "%s/%s", pwent->pw_dir, LOCATIONS);

    if((file = fopen(filename, "r")) == (FILE *)NULL) {
	fprintf(stderr, "Can't open save file %s\n", filename);
	perror(Myname);
	return;
    }

    while(fgets(buffer, sizeof(buffer), file)) {
	/* skip over leading spaces */
	for(cptr = buffer ; *cptr && isspace(*cptr) ; cptr++) ;
	
	if(sscanf(cptr, "%s %d %d", hostname, &x, &y) == 3
	   && (hostptr = nametohost(hostname))) {
	    XtMoveWidget(hostptr->hl_widget, (Position)x, (Position)y);
	}
    }
    fclose(file);
    
    return;
}

char *fullhostname(hostname) char *hostname; {
    static char		fullname[100];

    if(index(hostname, '.') == NULL) {
	sprintf(fullname, "%s.%s", hostname, _res.defdname);
	return(fullname);
    } else
	return(hostname);
}

PARSEHOST *newparsehost(name, incl) char *name; int incl; {
    register PARSEHOST	*pptr;

    if((pptr = (PARSEHOST *)malloc(sizeof(PARSEHOST)))
       == (PARSEHOST *)NULL) {
	fprintf(stderr, "Can't alloc %d bytes of memory! (newparsehost)\n"
		, sizeof(PARSEHOST));
	return((PARSEHOST *)NULL);
    }

    pptr->pah_name = cpystr(name);
    pptr->pah_display = (char *)NULL;
    pptr->pah_mask = DEFAULTNETMASK;
    pptr->pah_interval = DEFAULTPINGTIME;
    pptr->pah_community = cpystr(DEFAULTCOMMUNITY);
    pptr->pah_x = INITIAL_X;
    pptr->pah_y = INITIAL_Y;
    pptr->pah_flags = incl;	/* specify if we're in an include file */

    return(pptr);
}

freeparsehost(pptr) PARSEHOST *pptr; {

    if(pptr->pah_name) free(pptr->pah_name);
    if(pptr->pah_display) free(pptr->pah_display);
    if(pptr->pah_community) free(pptr->pah_community);

    free(pptr);

    return;
}

movehost(hptr, x, y) HOSTENTRY *hptr; short x, y; {

    if(! hptr) return;

    XtMoveWidget(hptr->hl_widget, (Position)x, (Position)y);
}

inhostlist(host) HOSTENTRY *host; {
    register HOSTENTRY	*hptr;

    for(hptr = Hostlist ; hptr && hptr != host ; hptr = hptr->hl_next) ;

    if(hptr)
	return(TRUE);
    else
	return(FALSE);
}
