#include "xsnmp.h"

#define Debug	(Flags & (PROG_DEBUG_IF | PROG_DEBUG))

/*
 * Interface stuff
 */
IFENTRY			*Ifentries=(IFENTRY *)NULL;

/*
 * Create a new interface entry
 */
IFENTRY *newifentry(hostaddr, hostmask) u_long hostaddr, hostmask; {
    register IFENTRY	*ifptr;
    
    if((ifptr = (IFENTRY *)malloc(sizeof(IFENTRY))) != (IFENTRY *)NULL) {
	ifptr->if_addr.s_addr = hostaddr;
	ifptr->if_net.s_addr = hostaddr & hostmask;
	ifptr->if_mask.s_addr = hostmask;
	ifptr->if_nextif = Ifentries;
	ifptr->if_next = (IFENTRY *)NULL;
	ifptr->if_flags = 0;
	ifptr->if_host = (HOSTENTRY *)NULL;
	ifptr->if_speed = 0;	/* unknown */
	ifptr->if_index = 0177;	/* unknown */
	Ifentries = ifptr;
	return(ifptr);
    }

    fprintf(stderr, "Can't alloc %d bytes of memory in newifentry!\n"
	    , sizeof(IFENTRY));

    return((IFENTRY *)NULL);
}

/*
 * Add a new interface for this host
 */
IFENTRY *addif(hostptr, hostaddr, hostmask)
HOSTENTRY	*hostptr;
u_long		hostaddr, hostmask; {
    register IFENTRY	**ifptr;
    struct in_addr	addr;
    
    if(Debug) {
	addr.s_addr = hostaddr;
	printf("Adding interface %s (mask 0x%x) for host %s\n"
	       , inet_ntoa(addr), hostmask, hostptr->hl_name);
    }

    for(ifptr = &hostptr->hl_if ; *ifptr ; ifptr = &(*ifptr)->if_next)
	if((*ifptr)->if_addr.s_addr == hostaddr)
	    return *ifptr;	/* there already */

    /* Add this entry */
    if((*ifptr = newifentry(hostaddr, hostmask)) == (IFENTRY *)NULL)
	return((IFENTRY *)NULL);

    (*ifptr)->if_host = hostptr; /* back pointer to this if's HOSTENTRY */

    /* And add it to the ping list */
    if(hostaddr != LOCALHOST)
	insertping(*ifptr, hostptr);

    return(*ifptr);
}

/*
 * See if this host has this interface address
 */
findif(hostptr, address) HOSTENTRY *hostptr; u_long address; {
    register IFENTRY	*ifptr;

    for(ifptr = hostptr->hl_if ; ifptr ; ifptr = ifptr->if_next)
	if(ifptr->if_addr.s_addr == address) return TRUE;

    return FALSE;
}

HOSTENTRY *iftohost(address) u_long address; {
    register HOSTENTRY	*hptr;
    register IFENTRY	*ifptr;
    
    for(hptr = Hostlist ; hptr ; hptr = hptr->hl_next)
	for(ifptr = hptr->hl_if ; ifptr ; ifptr = ifptr->if_next)
	    if(ifptr->if_addr.s_addr == address)
		return(hptr);

    return((HOSTENTRY *)NULL);
}

IFENTRY *addifmask(hptr, address, mask)
HOSTENTRY	*hptr;
u_long		address, mask; {
    register IFENTRY	*ifptr;
    struct in_addr	addr;

    if(! hptr) return((IFENTRY *)NULL);

    if(Debug) {
	addr.s_addr = address;
	printf("adding mask 0x%x to interface %s for host %s\n"
	       , mask, inet_ntoa(addr), hptr->hl_name);
    }

    for(ifptr = hptr->hl_if ; ifptr ; ifptr = ifptr->if_next)
	if(ifptr->if_addr.s_addr == address) {
	    ifptr->if_net.s_addr = address & mask;
	    ifptr->if_mask.s_addr = mask;
	    Flags |= PROG_KICKNET; /* redraw the network sometime */
	    break;
	}

    return(ifptr);
}

IFENTRY *addifindex(address, index) u_long address; int index; {
    register IFENTRY	*ifptr;

    for(ifptr = Ifentries ; ifptr ; ifptr = ifptr->if_nextif)
	if(ifptr->if_addr.s_addr == address && ifptr->if_index == 0177) {
	    ifptr->if_index = (char)(index & 0177);
	    break;
	}

    return(ifptr);
}

IFENTRY *addifindspeed(hptr, index, speed)
HOSTENTRY	*hptr;
int		index;
int		speed; {
    IFENTRY	*ifptr;

    for(ifptr = hptr->hl_if ; ifptr ; ifptr = ifptr->if_next)
	if(ifptr->if_index == index) {
	    if(ifptr->if_speed == 0) {
		ifptr->if_speed = speed;
		Flags |= PROG_KICKNET; /* redisplay network */
	    }
	}

    return(ifptr);
}

IFENTRY *setifindstatus(hptr, index, value)
HOSTENTRY	*hptr;
int		index, value; {
    IFENTRY	*ifptr;

    for(ifptr = hptr->hl_if ; ifptr ; ifptr = ifptr->if_next)
	if(ifptr->if_index == index) {
	    if(value == 1)
		ifptr->if_flags |= FL_UP;
	    else
		ifptr->if_flags |= FL_DOWN;
	    /* set this as the last status checked time */
	    ifptr->if_lastup = time(0L);
	}

    return(ifptr);
}

IFENTRY *ifindextoif(hptr, index) HOSTENTRY *hptr; int index; {
    register IFENTRY	*ifptr;

    for(ifptr = hptr->hl_if ; ifptr ; ifptr = ifptr->if_next)
	if(ifptr->if_index == index)
	    break;

    return(ifptr);
}

countif(hptr) HOSTENTRY *hptr; {
    register IFENTRY	*ifptr;
    register int	count;
    
    for(ifptr = hptr->hl_if , count = 0
	; ifptr
	; ifptr = ifptr->if_next , count++) ;

    return(count);
}

freeif(ifptr) IFENTRY *ifptr; {

    /* Make sure no ping entries point to it first */
    removepings(ifptr);

    free(ifptr);

    return;
}

removeif(ifentry) IFENTRY *ifentry; {
    register IFENTRY	*ifptr;
    
    if(ifentry == Ifentries)
	Ifentries = ifentry->if_nextif;
    else {
	for(ifptr = Ifentries
	    ; ifptr->if_nextif && ifptr->if_nextif != ifentry
	    ; ifptr = ifptr->if_nextif) ;

	/* found it? */
	if(ifptr->if_nextif == ifentry) {
	    ifptr->if_nextif = ifptr->if_nextif->if_nextif;
	}
    }

    freeif(ifentry);

    return;
}

removeifs(hptr) HOSTENTRY *hptr; {
    register IFENTRY	*ifptr;
    register IFENTRY	*nextif;

    for(ifptr = hptr->hl_if ; ifptr ; ifptr = nextif) {
	nextif = ifptr->if_next; /* save for use next */
	removeif(ifptr);
    }

    return;
}

PARSEIF *newparseif(name, incl) char *name; int incl; {
    register PARSEIF	*pptr;

    if((pptr = (PARSEIF *)malloc(sizeof(PARSEIF)))
       == (PARSEIF *)NULL) {
	fprintf(stderr, "Can't alloc %d bytes of memory! (newparseif)\n"
		, sizeof(PARSEIF));
	return((PARSEIF *)NULL);
    }

    pptr->pai_name = cpystr(name);
    pptr->pai_addr = 0;
    pptr->pai_mask = DEFAULTNETMASK;
    pptr->pai_flags = incl;

    return(pptr);
}

freeparseif(pptr) PARSEIF *pptr; {

    if(pptr->pai_name) free(pptr->pai_name);

    free(pptr);

    return;
}
