/*
 * Simple little program to ping as many systems as we can in regular
 * intervals.
 *
 * Author:	Richard Johnson		<raj@ncgia.ucsb.edu>
 */
#include "xsnmp.h"
#include <sys/stat.h>

#define Debug	(Flags & (PROG_DEBUG_MAIN | PROG_DEBUG))

char *Version="$Author: raj $ $Date: 93/06/29 17:11:46 $ - $Revision: 2.2 $";

int	Minsleep;
/* int	Debug=FALSE; */
int	Flags=0;		/* Program flags used in lots of places */
char	*Myname;		/* My invocation name */

FILE	*Xsnmprcfile;		/* file desc. for ~/.xsnmprc file */
char	*Xsnmprcp;		/* char ptr to input buffer */
char	*Xsnmprcname;		/* name of current included file */

RESOURCES	Resources;

static XtResource	resources[]={
    {"geometry",	"Geometry",	XtRString
	 , sizeof(char *), XtOffsetOf(RESOURCES, geometry), XtRString
	 , NULL}
};

struct {
    char	flg_char;
    int		flg_value;
}		Debugflags[]={
    {'x', PROG_DEBUG_X},	/* debug X routines */
    {'s', PROG_DEBUG_SNMP},	/* debug snmp routines */
    {'i', PROG_DEBUG_IF},	/* debug if routines */
    {'n', PROG_DEBUG_NET},	/* debug net routines */
    {'h', PROG_DEBUG_HOST},	/* debug host routines */
    {'d', PROG_DEBUG_DNS},	/* debug dns routines */
    {'f', PROG_DEBUG_FINGER},	/* debug finger routines */
    {'p', PROG_DEBUG_PING},	/* debug ping routines */
    {'m', PROG_DEBUG_MAIN},	/* debug main routines */
    {'y', PROG_DEBUG_YACC},	/* debug yacc parsing code */
};

/*
 * Do pings for as many systems as we should, then go back to sleep
 */
XtTimerCallbackProc alarmping() {
    register PINGENTRY	*pingent, *nextping;
    register int	i, j;
    int			count, pingtime, interval;
    IFENTRY		*ifptr;
    static u_long	lasttime=0;
    u_long		thistime;
    HOSTENTRY		*hostptr;
    int			can_exit;
    int			scoot;
    PINGENTRY		*pent;

    if(Debug)
	printf("Alarmping called:\n");

    thistime = time(0L);

    can_exit = FALSE;

    if(Pinglist) {
	if(Debug)
	    printf("%s is at front of list (interval=%d)\n"
		   , Pinglist->pen_if->if_host->hl_name
		   , Pinglist->pen_interval);
	    pingtime = thistime - lasttime;
	if(Debug)
	    printf("\t\tPingtime is %d\n", pingtime);
	/*
	 * Scoot everything down by pingtime seconds
	 */
	scoot = pingtime;
	for(pent = Pinglist
	    ; pent && scoot > 0
	    ; pent = pent->pen_next) {
	    if(pent->pen_interval == 0) continue;
	    if(Debug)
		printf("Adjusting %s from %d to %d\n"
		       , pent->pen_if->if_host->hl_name
		       , pent->pen_interval
		       , pent->pen_interval - scoot);
	    if((pent->pen_interval -= scoot) < 0) {
		if(Debug)
		    printf("Setting %s to 0\n"
			   , pent->pen_if->if_host->hl_name);
		scoot = -pent->pen_interval;
		pent->pen_interval = 0;
	    } else scoot = 0;
	}

	if(Debug) {
	    printf("Pinging hosts: "); fflush(stdout);
	}
	for(count = 0 , pingent = Pinglist , interval = 0
	    ; pingent && (interval + pingent->pen_interval) <= pingtime
	    ; pingent = pingent->pen_next) {
	    int			a;
#ifdef	sun
	    struct timeval	delaytime;
#endif
	    hostptr = pingent->pen_if->if_host;
	    if(Debug) {
		printf("%s(%d) "
		       , pingent->pen_if->if_host->hl_name
		       , pingent->pen_interval);
		fflush(stdout);
	    }
/*
 * I've found that on Sun workstations you have to put in a little delay here
 * or else you end up sending out packets too fast one after another.  A sun
 * sending a packet doesn't seem to be able to receive one at the same time.
 * (Maybe this is only a problem if it's sending and receiving on the same
 * port at the same time?  I don't know.)  Anyway, this little delay is all
 * that's needed to make it work.  (On a Sparc10 under SunOS 4.1.3 at least.)
 */
#ifdef	sun
	    delaytime.tv_sec = 0;
	    delaytime.tv_usec = 10;
	    select(0, &a, 0, 0, &delaytime);
#endif

	    if(pingent->pen_pings < MAXPINGS
	       || (pingent->pen_pings % PINGBACKOFF) == 0) {
		sendping(pingent->pen_if->if_addr.s_addr);
		if(pingent->pen_if->if_host->hl_flags & FL_SHOWPINGS)
		    flashhost(pingent);
	    }

	    interval += pingent->pen_interval;
	    /*
	     * Set it's interval to 0 so we'll ping it again if it doesn't
	     * respond within the sleep interval.
	     * (If it *does* respond, reinsertping will reset this.)
	     */
	    pingent->pen_interval = 0;

	    pingent->pen_pings++;
	}

/*	if(Debug) {
/*	    printf("\nRest of the queue is:");
/*	    for(pent = pingent ; pent ; pent = pent->pen_next)
/*		printf("%s(%d) "
/*		       , pent->pen_if->if_host->hl_name
/*		       , pent->pen_interval);
/*	    printf("\n");
/*	} */
	
	lasttime = thistime;
	
	if(pingent) {
	    /* Now call setup to call us for the next update */
	    if(Debug) {
		printf("%s is next with interval %d\n"
		       , pingent->pen_if->if_host->hl_name
		       , pingent->pen_interval);
		fflush(stdout);
	    }
	    if(pingent->pen_interval > 0
	       && pingent->pen_interval <= Minsleep) {
		if(Debug)
		    printf("Sleeping for %d seconds\n", pingent->pen_interval);
		XtAppAddTimeOut(XtWidgetToApplicationContext(Toplevel)
				, (u_long)(pingent->pen_interval * 1000)
				, (XtTimerCallbackProc)alarmping
				, (XtPointer)NULL);
		
		/*	alarm(Pinglist->pen_interval); */
	    } else /* alarm(1); */ {
		if(Debug)
		    printf("Sleeping for %d (min1) seconds\n", Minsleep);
		XtAppAddTimeOut(XtWidgetToApplicationContext(Toplevel)
				, Minsleep * 1000
				, (XtTimerCallbackProc)alarmping
				, (XtPointer)NULL);
	    }
	} else {
	    if(Debug)
		printf("Sleeping for %d (min2) seconds\n", Minsleep);
	    XtAppAddTimeOut(XtWidgetToApplicationContext(Toplevel)
			    , Minsleep * 1000
			    , (XtTimerCallbackProc)alarmping
			    , (XtPointer)NULL);
	}
    } else {
	if(Debug)
	    printf("Sleeping for %d (min3) seconds\n", Minsleep);
	XtAppAddTimeOut(XtWidgetToApplicationContext(Toplevel)
			, Minsleep * 1000
			, (XtTimerCallbackProc)alarmping
			, (XtPointer)NULL);
    }

    /* See if we need to redraw the network */
    if(Flags & PROG_KICKNET) drawnets();

    return;
}

getchildren() {
    union wait		*status;

    while(wait3(&status, WNOHANG, NULL) > 0) ;
}

main(argc, argv) int argc; char **argv; {
    register HOSTENTRY	*hptr;
    int			i;
    PINGENTRY		*pingent;
    IFENTRY		*ifptr;
    Widget		widget;

    if((Myname = (char *)rindex(argv[0], '/')) == (char *)NULL)
	Myname = argv[0];
    else
	Myname++;

    initpingsocket();
    (void) setuid(getuid());

    Toplevel = XtInitialize(argv[0], "Xsnmp", NULL, 0, &argc, argv);
    
    XtGetApplicationResources(Toplevel, (XtPointer)&Resources, resources
			      , XtNumber(resources), NULL, (Cardinal)0);

    for(--argc , ++argv ; *argv ; argv++ , argc--)
	if(**argv == '-')
	    switch( *++*argv ) { /* parse flags */
	    case 'D':		/* parse Debug flags */
		if(*++*argv == '\0') Flags |= PROG_DEBUG; /* everything */
		while(**argv) {
		    for(i = 0 ; i < ARRAYSIZE(Debugflags) ; i++) {
			if(Debugflags[i].flg_char == **argv) {
			    Flags |= Debugflags[i].flg_value;
			    break;
			}
		    }
		    if(i >= ARRAYSIZE(Debugflags))
			fprintf(stderr, "Debug letter '%c' not understood"
				, **argv);
		    ++*argv;
		}
		break;
	    case 'r':		/* reverse video */
		Flags |= PROG_REVERSE;
		while(*++*argv) ;
		--*argv;
		break;
	    }
    
    Mydisplay = XtDisplay(Toplevel);
    Myscreen = XDefaultScreen(Mydisplay);
    Myscrptr = XtScreen(Toplevel);
    
    /* Make sure we gather all of our children */
    signal(SIGCHLD, getchildren);

    /* Do we have color? */
    if(DefaultDepth(Mydisplay, Myscreen) > 1) Flags |= PROG_COLOR;

    /* Make up all of our menus */
    makehostmenu();
    makesnmpmenu();
    makenetmenu();
    makemainmenu();

    xt_addactions();		/* call our routine to do the addactions */

    XtSetMappedWhenManaged(Toplevel, False);

    XtRealizeWidget(Toplevel);

    i = 0;
    XtSetArg(Args[i], XtNwidth, &Formwidth); i++;
    XtSetArg(Args[i], XtNheight, &Formheight); i++;
    XtGetValues(Toplevel, Args, i);

    if(Debug)
	printf("Canvas width=%d, height=%d\n", Formwidth, Formheight);

    i = 0;
    XtSetArg(Args[i], XtNwidth, Formwidth); i++;
    XtSetArg(Args[i], XtNheight, Formheight); i++;
    if(Flags & PROG_REVERSE) {
	XtSetArg(Args[i], XtNbackground, BlackPixelOfScreen(Myscrptr)); i++;
	XtSetArg(Args[i], XtNforeground, WhitePixelOfScreen(Myscrptr)); i++;
    } else {
	XtSetArg(Args[i], XtNforeground, BlackPixelOfScreen(Myscrptr)); i++;
	XtSetArg(Args[i], XtNbackground, WhitePixelOfScreen(Myscrptr)); i++;
    }
    Form = XtCreateManagedWidget("canvas", compositeWidgetClass, Toplevel
				 , Args, i);

    XtAddEventHandler(Form, StructureNotifyMask, False
		      , (XtEventHandler)xt_resize, (XtPointer)0);

    Cmap = XDefaultColormap(Mydisplay, Myscreen);
    XParseColor(Mydisplay, Cmap, "Black", &Black);
    XParseColor(Mydisplay, Cmap, "White", &White);
    if(Flags & PROG_COLOR)
	XParseColor(Mydisplay, Cmap, "Gray", &Gray);

    i = 0;
    Exit = XtCreateManagedWidget("exit", commandWidgetClass, Form
				       , Args, i);
    XtAddCallback(Exit, "callback", xt_exitproc, 0);

    if(readrcfile())
	parserc();
    else {
	widget = readhosts(Exit);
	/* Set host locations on the screen */
	getlocations();
    }

    /* Create the background pixmap */
    Lines = XCreatePixmap(Mydisplay, XtWindow(Toplevel)
			  , Formwidth, Formheight
			  , DefaultDepthOfScreen(Myscrptr));
    Gc = XCreateGC(Mydisplay, XtWindow(Toplevel), NULL, NULL);
    if(Flags & PROG_REVERSE)
	XSetForeground(Mydisplay, Gc, BlackPixelOfScreen(Myscrptr));
    else
	XSetForeground(Mydisplay, Gc, WhitePixelOfScreen(Myscrptr));
    XFillRectangle(Mydisplay, Lines, Gc, 0, 0, Formwidth, Formheight);
    if(Flags & PROG_REVERSE) {
	XSetBackground(Mydisplay, Gc, BlackPixelOfScreen(Myscrptr));
	XSetForeground(Mydisplay, Gc, WhitePixelOfScreen(Myscrptr));
    } else {
	XSetForeground(Mydisplay, Gc, BlackPixelOfScreen(Myscrptr));
	XSetBackground(Mydisplay, Gc, WhitePixelOfScreen(Myscrptr));
    }

    Minsleep = 1;

    XtMapWidget(Toplevel);

    initpings();
    
    /* Now we call our ping alarm routine to start it going */
    alarmping();
    
    snmphosts();

    snmpifs();

    drawnets();

    i = 0;
    XtSetArg(Args[i], XtNbackgroundPixmap, Lines); i++;
    XtSetValues(Form, Args, i);
    
/*    XtInstallAllAccelerators(Form, Form); */

    Flags |= PROG_WINDOW;	/* say we have a window now */

    XtMainLoop();
}

/*
 * Open and parse the ~/.xsnmprc file
 */
parserc() {
    struct passwd	*pwent;
    char		buffer[BUFSIZ];

    /* First find their home directory */
    if((pwent = getpwuid(getuid())) == (struct passwd *)NULL
       || (Xsnmprcname = (char *)malloc(strlen(pwent->pw_dir)
				      + sizeof(XSNMPRC) + 1))
       == (char *)NULL) {
	fprintf(stderr, "Error making name of ~/%s file\n", XSNMPRC);
	exit(1);
    }
    sprintf(Xsnmprcname, "%s/%s", pwent->pw_dir, XSNMPRC);

    if((Xsnmprcfile = fopen(Xsnmprcname, "r")) == (FILE *)NULL) {
	fprintf(stderr, "Can't open hostlist file %s\n", Xsnmprcname);
	perror(Myname);
	return;
    }
    
    Xsnmprcp = (char *)NULL;

    yyparse();
}

xsnmprcinput() {
    int		c;
    extern int	yylineno;
/*    static char	buffer[BUFSIZ];
/*
/*    if(Xsnmprcp == (char *)NULL || *Xsnmprcp == '\0') {
/*	if(fgets(buffer, sizeof(buffer), Xsnmprcfile) == (char *)NULL)
/*	    return(NULL);
/*	yylineno++;
/*	Xsnmprcp = buffer;
/*    }
/*    
/*    return((int)*Xsnmprcp++); */

    while((c = fgetc(Xsnmprcfile)) == EOF) {
	if(! popinclude())
	    return(NULL);
    }
    
    if(Debug)
	printf("<%c>", (char)c);

    if((c & 0277) == '\n') yylineno++;

    return(c);
}

xsnmprcunput(c) int c; {

    if((c & 0277) == '\n') yylineno--;
    ungetc((int)c, Xsnmprcfile);

/*    if(Xsnmprcp)
/*	*(--Xsnmprcp) = c; */
}

xsnmprcoutput(c) char c; {

    putchar(c);
}

/*
 * Determine if we should read the ~/.xsnmprc or ~/.xsnmp_hosts and locations
 * files.
 *
 * If the ~/.xsnmprc doesn't exist or if ~/.xsnmp_hosts is newer, then read
 * older style file, else read the newer.
 */
readrcfile() {
    struct passwd	*pwent;
    char		*filename;
    struct stat		statrc;
    struct stat		stathost;

    /* First find their home directory */
    if((pwent = getpwuid(getuid())) == (struct passwd *)NULL
       || (filename = (char *)malloc(strlen(pwent->pw_dir) + 80))
       == (char *)NULL) {
	fprintf(stderr, "Error in checking status of .xsnmp... files\n");
	exit(1);
    }
    sprintf(filename, "%s/%s", pwent->pw_dir, XSNMPRC);

    if(stat(filename, &statrc) < 0)
	return(FALSE);		/* no ~/.xsnmprc, read ~/.xsnmp_hosts */

    sprintf(filename, "%s/%s", pwent->pw_dir, HOSTLIST);

    if(stat(filename, &stathost) < 0)
	return(TRUE);		/* no ~/.xsnmp_hosts, read ~/.xsnmprc */

    if(statrc.st_mtime > stathost.st_mtime)
	return(TRUE);		/* ~/.xsnmprc is newer */
    else
	return(FALSE);		/* ~/.xsnmp_hosts is newer */
}
