#include "xsnmp.h"

#define Debug	(Flags & (PROG_DEBUG_X | PROG_DEBUG))

/*
 * X routines
 */

/* Globals */
Widget			Toplevel, Form;
Widget			Exit;	/* the "exit" widget */
Arg			Args[20];		/* used everywhere else! */
Colormap		Cmap;
XColor			White, Black;		/* colors */
Display			*Mydisplay;
int			Myscreen;
Screen			*Myscrptr;
Pixmap			Lines;
Dimension		Formwidth;
Dimension		Formheight;
GC			Gc;
Widget			Hostwidget;

/* Globals only to this file */
static Pixmap		graypixmap=(Pixmap)NULL;
static HOSTENTRY	*Hostptr;
static NETENTRY		*Netptr;
static int		Hostx, Hosty;

/* A useful routine */
#define SETFLOATARG(a, v, n)	{			\
	float _local_f=(n);				\
	    if(sizeof(float) > sizeof(XtArgVal)) {	\
		XtSetArg((a), (v), &(_local_f));	\
	    } else {					\
		XtArgVal *val=(XtArgVal *)&(_local_f);	\
		XtSetArg((a), (v), *val);		\
	    }						\
	}

XtTimerCallbackProc xt_exit2() {

    XtDestroyWidget(Toplevel);

    exit(0);
}

/*
 * Exit the program.
 */
void xt_exitproc(w) Widget w; {
    char	buf[80];

/*    XFreePixmap(Mydisplay, Lines);
/*    XFreePixmap(Mydisplay, graypixmap); */

    XtUnmapWidget(Toplevel);

    XtAppAddTimeOut(XtWidgetToApplicationContext(Toplevel)
		    , (u_long)(2 * 1000)
		    , (XtTimerCallbackProc)xt_exit2
		    , (XtPointer)0);
}

xt_remtopshell(w) Widget w; {

    while(XtClass(w) != applicationShellWidgetClass)
	w = XtParent(w);

    XtDestroyWidget(w);
}

Widget xt_promptline(form, labelname, editname, editstring, fromvert, edit)
Widget	form;
char	*labelname, *editname, *editstring;
Widget	fromvert;
XawTextEditType	edit; {
    register int	i;
    Widget		w;
    
    i = 0;
    XtSetArg(Args[i], XtNfromVert, fromvert); i++;
    XtSetArg(Args[i], XtNfromHoriz, (Widget)NULL); i++;
    w = XtCreateManagedWidget(labelname, labelWidgetClass, form
			      , Args, i);

    i = 0;
    XtSetArg(Args[i], XtNfromVert, fromvert); i++;
    XtSetArg(Args[i], XtNfromHoriz, w); i++;
    XtSetArg(Args[i], XtNstring, editstring); i++;
    XtSetArg(Args[i], XtNeditType, edit); i++;
    w = XtCreateManagedWidget(editname, asciiTextWidgetClass, form
			      , Args, i);
    return(w);
}

Widget real_disptext(name, buffer, dismiss) char *name, *buffer; int dismiss; {
    register int	i, j;
    register char	*cp, *cp2;
    Dimension		maxheight;
    Dimension		height;
    Dimension		width;
    int			myfontheight;
    Widget		shell, box, w;
    XFontStruct		*fontstr;

    if(buffer == (char *)NULL)
	buffer = " ";

    i = 0;
    XtSetArg(Args[i], XtNfont, &fontstr); i++;
    XtGetValues(Exit, Args, i);

    i = 0;
    shell = XtCreatePopupShell(name, applicationShellWidgetClass
			       , Form, Args, i);
    
    i = 0;
    box = XtCreateManagedWidget("textbox", boxWidgetClass, shell, Args, i);

    /* Calculate total box height - make sure it's not too big */
    /* get font height */
    myfontheight = fontstr->max_bounds.ascent + fontstr->max_bounds.descent;
    for(width = 0 , cp = buffer ; *cp ; cp++) {
	for(cp2 = cp ; *cp2 && *cp2 != '\n' ; cp2++) ;
	if((i = XTextWidth(fontstr, cp, cp2 - cp)) > width) width = i;
	cp = cp2;
    }

    for(height = myfontheight , cp = buffer ; *cp ; cp++)
	if(*cp == '\n' && *(cp+1)) height += myfontheight;
    maxheight = DisplayHeight(Mydisplay, Myscreen) * .75;

    i = 0;
    if(height > maxheight) height = maxheight;
    XtSetArg(Args[i], XtNheight, height+4); i++;
    XtSetArg(Args[i], XtNwidth, width + SCROLLBARWIDTH*2); i++;
    XtSetArg(Args[i], XtNstring, buffer); i++;
    XtSetArg(Args[i], XtNscrollVertical, XawtextScrollWhenNeeded); i++;
    w = XtCreateManagedWidget("textinfo", asciiTextWidgetClass, box, Args, i);

    if(dismiss) {
	i = 0;
	w = XtCreateManagedWidget("Dismiss", commandWidgetClass, box, Args, i);
	XtAddCallback(w, "callback"
		      , (XtCallbackProc)xt_remtopshell, (XtPointer)0);
    } else {
	XtSetArg(Args[0], XtNgeometry, "+0+0");
	XtSetValues(shell, Args, 1);
    }

    XtPopup(shell, XtGrabNone);

    return(shell);
}

Widget disptext(name, buffer) char *name, *buffer; {

    return(real_disptext(name, buffer, TRUE)); /* with a dismiss button */
}

/*
 * Open a file and save all information in old ~/.xsnmp_hosts style
 */
void xt_save_old() {
    struct Coord	start;
    Position		x, y;
    FILE		*locations, *hosts;
    char		buffer[80];
    char		*filename;
    struct passwd	*pwent;
    IFENTRY		*ifptr;
    HOSTENTRY		*hptr;

    /* First find their home directory */
    if((pwent = getpwuid(getuid())) == (struct passwd *)NULL
       || (filename = (char *)malloc(strlen(pwent->pw_dir)
				      + sizeof(LOCATIONS) + 1))
       == (char *)NULL) {
	sprintf(buffer, "Error making name of ~/%s file\n", LOCATIONS);
	disptext("Save Error", buffer);
	return;
    }
    sprintf(filename, "%s/%s", pwent->pw_dir, LOCATIONS);

    if((locations = fopen(filename, "w")) == (FILE *)NULL) {
	sprintf(buffer, "Can't open save file %s\n", filename);
	disptext("Save Error", buffer);
	perror("openfile");
	return;
    }

    /* First find their home directory */
    free(filename);
    if((filename = (char *)malloc(strlen(pwent->pw_dir)
				      + sizeof(HOSTLIST) + 1))
       == (char *)NULL) {
	sprintf(buffer, "Error making name of ~/%s file\n", HOSTLIST);
	disptext("Save Error", buffer);
	return;
    }
    sprintf(filename, "%s/%s", pwent->pw_dir, HOSTLIST);

    if((hosts = fopen(filename, "w")) == (FILE *)NULL) {
	sprintf(buffer, "Can't open save file %s\n", filename);
	disptext("Save Error", buffer);
	perror("openfile");
	return;
    }
    free(filename);

    for(hptr = Hostlist ; hptr ; hptr = hptr->hl_next) {
	getposition(hptr->hl_widget, &x, &y);
	fprintf(locations, "%s %d %d\n", hptr->hl_name, x, y);
	writehostinfo(hosts, hptr);
    }

    fclose(locations);
    fclose(hosts);
}

/*
 * Open ~/.xsnmprc and save all information (new style)
 */
void xt_save() {
    struct Coord	start;
    Position		x, y;
    FILE		*xsnmprc;
    char		buffer[80];
    char		*filename;
    struct passwd	*pwent;
    IFENTRY		*ifptr;
    HOSTENTRY		*hptr;
    NETENTRY		*nptr;
    
    /* First find their home directory */
    if((pwent = getpwuid(getuid())) == (struct passwd *)NULL
       || (filename = (char *)malloc(strlen(pwent->pw_dir)
				     + sizeof(XSNMPRC) + 1))
       == (char *)NULL) {
	sprintf(buffer, "Error making name of ~/%s file\n", XSNMPRC);
	disptext("Save Error", buffer);
	return;
    }
    sprintf(filename, "%s/%s", pwent->pw_dir, XSNMPRC);
    
    if((xsnmprc = fopen(filename, "w")) == (FILE *)NULL) {
	sprintf(buffer, "Can't open save file %s\n", filename);
	disptext("Save Error", buffer);
	perror("openfile");
	return;
    }
    
    /* Write out all host info */
    for(hptr = Hostlist ; hptr ; hptr = hptr->hl_next) {
	if(hptr->hl_flags & FL_INCLUDE) continue; /* from include file */
	getposition(hptr->hl_widget, &x, &y);
	writehostrcinfo(xsnmprc, hptr, x, y);
    }
    
    /* Write all network info */
    for(nptr = Netlist ; nptr ; nptr = nptr->nl_next)
	writenetrcinfo(xsnmprc, nptr);
    
    /* Write all include info */
    writeincludercinfo(xsnmprc);

    fclose(xsnmprc);
}

writehostrcinfo(xsnmprc, hptr, x, y)
FILE		*xsnmprc;
HOSTENTRY	*hptr;
Position	x, y; {
    register IFENTRY	*ifptr;

    if(hptr->hl_flags & FL_INCLUDE) return;

    fprintf(xsnmprc, "host \"%s\" loc %d %d ", hptr->hl_name, x, y);
    if(sstrcmp(hptr->hl_displayname, hptr->hl_name) != 0)
	fprintf(xsnmprc, "display \"%s\" ", hptr->hl_displayname);
    if(hptr->hl_if->if_mask.s_addr != DEFAULTNETMASK)
	fprintf(xsnmprc, "mask 0x%x ", hptr->hl_if->if_mask.s_addr);
    if(hptr->hl_interval != DEFAULTPINGTIME)
	fprintf(xsnmprc, "interval %d ", hptr->hl_interval);
    if(sstrcmp(hptr->hl_community, DEFAULTCOMMUNITY) != 0)
	fprintf(xsnmprc, "community \"%s\" ", hptr->hl_community);

    fprintf(xsnmprc, "\n");

    for(ifptr = hptr->hl_if->if_next ; ifptr ; ifptr = ifptr->if_next) {
	if(! (ifptr->if_flags & FL_SNMP)
	   && ifptr->if_addr.s_addr != LOCALHOST) 
	    writeifrcinfo(xsnmprc, ifptr);
    }
}

writeifrcinfo(file, ifptr) FILE *file; IFENTRY *ifptr; {
    register HOSTENTRY	*hptr;
    
    if(ifptr->if_flags & FL_INCLUDE) return;

    if((hptr = ifptr->if_host) == (HOSTENTRY *)NULL) return;
    
    fprintf(file, "interface \"%s\" ", hptr->hl_name);
    fprintf(file, "ipaddr %s ", inet_ntoa(ifptr->if_addr));
    if(ifptr->if_mask.s_addr != DEFAULTNETMASK)
	fprintf(file, "mask 0x%x ", ifptr->if_mask.s_addr);
    
    fprintf(file, "\n");
}

writenetrcinfo(file, nptr) FILE *file; NETENTRY *nptr; {
    
    if(nptr->nl_flags & NET_INCLUDE) return;

    fprintf(file, "net %s ", inet_ntoa(nptr->nl_net));
    if(nptr->nl_name && *nptr->nl_name)
	fprintf(file, "name \"%s\" ", nptr->nl_name);
    
    if(nptr->nl_flags & NET_CIRCLE) {
	if(nptr->nl_flags & NET_PINNED) {
	    fprintf(file, "center %d %d ", nptr->nl_startx
		    , nptr->nl_starty);
	} else
	    fprintf(file, "type circle ");
	if(nptr->nl_radius != DEFAULTRADIUS)
	    fprintf(file, "radius %d ", nptr->nl_radius);
    } else {			/* line */
	if(nptr->nl_flags & NET_PINNED) {
	    fprintf(file, "startloc %d %d endloc %d %d "
		    , nptr->nl_startx, nptr->nl_starty
		    , nptr->nl_endx, nptr->nl_endy);
	} else
	    fprintf(file, "type line ");
    }
    
    if(nptr->nl_flags & NET_SPEEDSET)
	fprintf(file, "speed %d ", nptr->nl_speed);

    fprintf(file, "\n");
}

writehostinfo(file, hptr) FILE *file; HOSTENTRY *hptr; {
    register IFENTRY	*ifptr;
    u_long		mask;
    int			interval;
    char		*community;

    if((mask = hptr->hl_if->if_mask.s_addr) == DEFAULTNETMASK)
	mask = 0;
    if((interval = hptr->hl_interval) == DEFAULTPINGTIME)
	interval = 0;
    if(sstrcmp((community = hptr->hl_community), DEFAULTCOMMUNITY) == 0)
	community = "";

    fprintf(file, "%s:%x:%d:%s\n", hptr->hl_displayname
	    , mask, interval, community);
    for(ifptr = hptr->hl_if->if_next ; ifptr ; ifptr = ifptr->if_next) {
	if(! (ifptr->if_flags & FL_SNMP)
	   && ifptr->if_addr.s_addr != LOCALHOST) 
	    fprintf(file, "+%s\n", inet_ntoa(ifptr->if_addr));
    }
}

XtEventHandler xt_move_widget(w, widget, event)
Widget		w, widget;
XButtonEvent	*event; {
    register Position	x, y;
    
    x = ((event->x + 5) / 10) * 10;
    y = ((event->y + 5) / 10) * 10;
    XtMoveWidget(widget, x, y);
    XtRealizeWidget(Form);
    XtRemoveEventHandler(Form, ButtonPressMask, False
			 , (XtEventHandler)xt_move_widget
			 , (XtPointer)widget);
    XtAddEventHandler(Form, ButtonPressMask, False
		      , (XtEventHandler)xt_save, (XtPointer)0);
    XtUngrabPointer(Form, CurrentTime);

    drawnets();
}

void xt_movehost(w, event) Widget w; XEvent *event; {
    register Position	x, y;
    static Position	wx=0, wy=0;
    
/*    printf("Event type %d in window 0x%x at %d, %d\n", event->type
/*	   , ((XButtonEvent *)event)->window
/*	   , ((XButtonEvent *)event)->x_root
/*	   , ((XButtonEvent *)event)->y_root); */
	   
/*    x = ((((XButtonEvent *)event)->x + 5) / 10) * 10;
/*    y = ((((XButtonEvent *)event)->y + 5) / 10) * 10; */
    x = ((XButtonEvent *)event)->x_root;
    y = ((XButtonEvent *)event)->y_root;
    
    if(!Hostwidget) return;

    if(Hostx == 0) {
	getposition(Hostwidget, &wx, &wy);
	Hostx = x - wx;
	Hosty = y - wy;
    }

    XtMoveWidget(Hostwidget
		 , (Position)((x - Hostx))
		 , (Position)((y - Hosty)));
/*		 , (Position)(x - Hostptr->hl_width/2)
/*		 , (Position)(y - Hostptr->hl_height/2)); */
/*    printf("Moved to %d, %d\n"
/*	   ,(x-Hostx), (y-Hosty));
/*		 , x - Hostptr->hl_width/2
/*		 , y - Hostptr->hl_height/2); */
}

void xt_gridhost(w) Widget w; {
    Position	x, y;

    getposition(Hostwidget, &x, &y);

    x = ((x + 5) / 10) * 10;
    y = ((y + 5) / 10) * 10;

    XtMoveWidget(Hostwidget, x, y);
}

gridhost(hptr) HOSTENTRY *hptr; {

    Hostwidget = hptr->hl_widget;

    xt_gridhost(Hostwidget);
}

void xt_menumove(w) Widget w; {
    register Widget	widget;
    
    xt_move(Hostwidget);
}

void xt_selecthost(w, event) Widget w; XEvent *event; {
    register HOSTENTRY	*hostptr;
    
    if((hostptr = widgettohost(w)) == (HOSTENTRY *)NULL) return;

    Hostwidget = w;
    Hostptr = hostptr;
    Hostx = 0; Hosty = 0;
}

xt_move(w) Widget w; {
    int		x, y, i;
    
    XtGrabPointer(Form, False, ButtonPressMask, GrabModeAsync, GrabModeAsync
		  , XtWindow(Form), None, CurrentTime);
 
    XtRemoveEventHandler(Form, ButtonPressMask, False, xt_save, 0);
    XtAddEventHandler(Form, ButtonPressMask, False
		      , (XtEventHandler)xt_move_widget
		      , (XtPointer)w);
    
/*    XDrawLine(Mydisplay, Lines, Gc, 0, 0, x, y);
/*    XtSetArg(Args[0], XtNbackgroundPixmap, Lines);
/*    XtSetValues(Form, Args, 1);
/*    XClearArea(Mydisplay, XtWindow(Form), 0, 0, 500,500, True); */
}

getposition(w, ax, ay) Widget w; Position *ax, *ay; {
    register int	i;
    
    i = 0;
    XtSetArg(Args[i], XtNx, ax), i++;
    XtSetArg(Args[i], XtNy, ay), i++;
    XtGetValues(w, Args, i);
}

/*
 * Make this host gray
 */
setgray(hostptr) HOSTENTRY *hostptr; {
    register int	i;
    
    if(Flags & PROG_COLOR) {
	i = 0;
	XtSetArg(Args[i], XtNbackground, Gray.pixel), i++;
	XtSetValues(hostptr->hl_widget, Args, i);
	return;
    }
    
    /* Normal case for monochrome */
    if(! graypixmap)
	graypixmap = XCreatePixmapFromBitmapData(Mydisplay, XtWindow(Toplevel)
						 , gray_bits, gray_width
						 , gray_height
						 , BlackPixelOfScreen(Myscrptr)
						 , WhitePixelOfScreen(Myscrptr)
				      , DefaultDepthOfScreen(Myscrptr));
    
    i = 0;
    XtSetArg(Args[i], XtNbackgroundPixmap, graypixmap); i++;

    XtSetValues(hostptr->hl_widget, Args, i);
}

setblack(hostptr) HOSTENTRY *hostptr; {
    register int	i;
    
    i = 0;
    XtSetArg(Args[i], XtNbackground, BlackPixelOfScreen(Myscrptr)), i++;
    XtSetArg(Args[i], XtNforeground, WhitePixelOfScreen(Myscrptr)), i++;
    XtSetArg(Args[i], XtNbackgroundPixmap, XtUnspecifiedPixmap), i++;
    XtSetValues(hostptr->hl_widget, Args, i);
}

setwhite(hostptr) HOSTENTRY *hostptr; {
    register int	i;
    
    i = 0;
    XtSetArg(Args[i], XtNforeground, BlackPixelOfScreen(Myscrptr)), i++;
    XtSetArg(Args[i], XtNbackground, WhitePixelOfScreen(Myscrptr)), i++;
    XtSetArg(Args[i], XtNbackgroundPixmap, XtUnspecifiedPixmap), i++;
    XtSetValues(hostptr->hl_widget, Args, i);
}

XtTimerCallbackProc xt_removenote(w) Widget w; {

    xt_remtopshell(w);
}

Widget dispnotice(name, buffer) char *name, *buffer; {
    register Widget	w;

    if(! (Flags & (PROG_WINDOW))) {
	fprintf(stderr, "%s: %s - %s\n", Myname, name, buffer);
	return;
    }

    w = real_disptext(name, buffer, FALSE); /* no dismiss button */

    XtAppAddTimeOut(XtWidgetToApplicationContext(Toplevel)
		    , (u_long)(NOTICEDISPTIME * 1000)
		    , (XtTimerCallbackProc)xt_remtopshell
		    , (XtPointer)w);
}

void xt_showinterfaces() {
    register HOSTENTRY	*hptr;
    register IFENTRY	*ifptr;
    char		buffer[80];
    char		*bufptr, *bufend;
    int			bufsize;
    char		*str1, *str2;
    char		derived[5];
    char		status;
    
    if((hptr = widgettohost(Hostwidget)) == (HOSTENTRY *)NULL) {
	return;			/* internal error? */
    }

    bufptr = (char *)NULL;
    bufend = (char *)NULL;
    bufsize = 0;

    appendstr("Derived Index Status    IP address      Network number  Speed\n"
	      , &bufptr, &bufsize, &bufend);

    for(ifptr = hptr->hl_if ; ifptr ; ifptr = ifptr->if_next) {
	str1 = cpystr(inet_ntoa(ifptr->if_addr));
	if(ifptr->if_net.s_addr == 0L) {
	    if(ifptr->if_flags & FL_SNMPMASK) str2 = cpystr("SNMP");
	    else if(ifptr->if_flags & FL_CONFMASK) str2 = cpystr("CONF");
	} else
	    str2 = cpystr(inet_ntoa(ifptr->if_net));
	derived[0] = '\0';
	status = '?';

	if(hptr->hl_flags & FL_INCLUDE)
	    strcat(derived, "I");

	if(ifptr->if_flags & FL_SNMP)
	    strcat(derived,"S");
	else if(ifptr->if_flags & FL_INCLUDE)
	    strcat(derived,"I");

	if(ifptr->if_flags & FL_UP) status = 'U';
	else if(ifptr->if_flags & FL_DOWN) status = 'D';

	if(ifptr->if_index == 0177)
	    sprintf(buffer, "%5s           %c  %15.15s     %15.15s  %5d\n"
		    , derived
		    , status, str1, str2
		    , ifptr->if_speed);
	else
	    sprintf(buffer, "%5s     %3d   %c  %15.15s     %15.15s  %5d\n"
		    , derived
		    , (int)ifptr->if_index, status
		    , str1, str2, ifptr->if_speed);
	free(str1); free(str2);
	appendstr(buffer, &bufptr, &bufsize, &bufend);
    }

    sprintf(buffer, "%s interfaces", hptr->hl_displayname);
    disptext(buffer, bufptr);
    free(bufptr);
}

void xt_showallpings() {
    HOSTENTRY	*hptr;

    for(hptr = Hostlist ; hptr ; hptr = hptr->hl_next)
	if(hptr->hl_flags & FL_SHOWPINGS)
	    hptr->hl_flags &= ~FL_SHOWPINGS;
	else
	    hptr->hl_flags |= FL_SHOWPINGS;
}

void xt_showpings() {
    register HOSTENTRY	*hptr;

    if((hptr = widgettohost(Hostwidget)) == (HOSTENTRY *)NULL) {
	return;			/* internal error? */
    }

    /* set the "show pings" flag */
    if(hptr->hl_flags & FL_SHOWPINGS)
	hptr->hl_flags &= ~FL_SHOWPINGS;
    else
	hptr->hl_flags |= FL_SHOWPINGS;
}

XtCallbackProc xt_addhost_do(w, topwidget) Widget w, topwidget; {
    register int	i;
    char		*hostname, *displayname, *netmask, *community;
    char		*pinginterval;
    u_long		hostnet;
    int			pingtime;
    HOSTENTRY		*hptr;

    /* Gather all info */
    i = 0;
    XtSetArg(Args[i], XtNstring, &hostname); i++;
    XtGetValues(XtNameToWidget(topwidget, "hostname"), Args, i);
    i = 0;
    XtSetArg(Args[i], XtNstring, &displayname); i++;
    XtGetValues(XtNameToWidget(topwidget, "displayname"), Args, i);
    i = 0;
    XtSetArg(Args[i], XtNstring, &netmask); i++;
    XtGetValues(XtNameToWidget(topwidget, "netmask"), Args, i);
    i = 0;
    XtSetArg(Args[i], XtNstring, &community); i++;
    XtGetValues(XtNameToWidget(topwidget, "community"), Args, i);
    i = 0;
    XtSetArg(Args[i], XtNstring, &pinginterval); i++;
    XtGetValues(XtNameToWidget(topwidget, "pinginterval"), Args, i);

    /* Change netmask and pinginterval into correct form */
    if(isdigit(*pinginterval))
	pingtime = atoi(pinginterval);
    else
	pingtime = DEFAULTPINGTIME;
    if(! *netmask)
	hostnet = DEFAULTNETMASK;
    else if(index(netmask, '.'))
	hostnet = inet_addr(netmask);
    else
	sscanf(netmask, "%x", &hostnet);
    if(*community == '\0')
	community = DEFAULTCOMMUNITY;
	
    if((hptr = addhost(hostname, hostnet, community, pingtime, displayname))
       != (HOSTENTRY *)NULL)
	addhosttonet(hptr);	/* move this host to a good place */
}

void xt_delhost() {
    register HOSTENTRY	*hptr;

    if((hptr = widgettohost(Hostwidget)) == (HOSTENTRY *)NULL)
	return;			/* internal error? */

    /* Remove this host from the host list */
    removehost(hptr);
}

void xt_addhost() {
    register int	i;
    Widget		hostform, shell, w;
    Widget		row, netmask;
    char		buffer[40];
    struct in_addr	addr;

    i = 0;
    shell = XtCreatePopupShell("Addhost form", applicationShellWidgetClass
			       , Toplevel, Args, i);
    i = 0;
    hostform = XtCreateManagedWidget("addhostform", formWidgetClass, shell
				     , Args, i);

    row = (Widget)NULL;

    /* Now we need to ask for hostname, netmask, community, pinginterval */
    i = 0;
    XtSetArg(Args[i], XtNfromVert, row); i++;
    XtSetArg(Args[i], XtNfromHoriz, row); i++;
    w = XtCreateManagedWidget("hostname-prompt", labelWidgetClass, hostform
			      , Args, i);

    i = 0;
    XtSetArg(Args[i], XtNfromVert, row); i++;
    XtSetArg(Args[i], XtNfromHoriz, w); i++;
    XtSetArg(Args[i], XtNeditType, XawtextEdit); i++;
    row = XtCreateManagedWidget("hostname", asciiTextWidgetClass, hostform
			      , Args, i);

    /* Ask for the display name as well */
    i = 0;
    XtSetArg(Args[i], XtNfromVert, row); i++;
    XtSetArg(Args[i], XtNfromHoriz, (Widget)NULL); i++;
    w = XtCreateManagedWidget("displayname-prompt", labelWidgetClass, hostform
			      , Args, i);

    i = 0;
    XtSetArg(Args[i], XtNfromVert, row); i++;
    XtSetArg(Args[i], XtNfromHoriz, w); i++;
    XtSetArg(Args[i], XtNeditType, XawtextEdit); i++;
    row = XtCreateManagedWidget("displayname", asciiTextWidgetClass, hostform
			      , Args, i);

    /* ask for netmask */
    i = 0;
    /* make sure this is wide enough for any netmask */
    XtSetArg(Args[i], XtNstring, "MMM.MMM.MMM.MMM"); i++;
    XtSetArg(Args[i], XtNfromVert, row); i++;
    XtSetArg(Args[i], XtNfromHoriz, (Widget)NULL); i++;
    w = XtCreateManagedWidget("netmask-prompt", labelWidgetClass, hostform
			      , Args, i);

    i = 0;
    XtSetArg(Args[i], XtNfromVert, row); i++;
    XtSetArg(Args[i], XtNfromHoriz, w); i++;
    XtSetArg(Args[i], XtNstring, buffer); i++;
    XtSetArg(Args[i], XtNeditType, XawtextEdit); i++;
    netmask = XtCreateManagedWidget("netmask", asciiTextWidgetClass, hostform
			      , Args, i);
    row = netmask;

    /* ask for community */
    i = 0;
    XtSetArg(Args[i], XtNfromVert, row); i++;
    XtSetArg(Args[i], XtNfromHoriz, (Widget)NULL); i++;
    w = XtCreateManagedWidget("community-prompt", labelWidgetClass, hostform
			      , Args, i);

    i = 0;
    XtSetArg(Args[i], XtNfromVert, row); i++;
    XtSetArg(Args[i], XtNfromHoriz, w); i++;
    XtSetArg(Args[i], XtNstring, DEFAULTCOMMUNITY); i++;
    XtSetArg(Args[i], XtNeditType, XawtextEdit); i++;
    row = XtCreateManagedWidget("community", asciiTextWidgetClass, hostform
			      , Args, i);

    /* ask for ping interval */
    i = 0;
    XtSetArg(Args[i], XtNfromVert, row); i++;
    XtSetArg(Args[i], XtNfromHoriz, (Widget)NULL); i++;
    w = XtCreateManagedWidget("ping-interval-prompt", labelWidgetClass
			      , hostform, Args, i);

    sprintf(buffer, "%d", DEFAULTPINGTIME);
    i = 0;
    XtSetArg(Args[i], XtNfromVert, row); i++;
    XtSetArg(Args[i], XtNfromHoriz, w); i++;
    XtSetArg(Args[i], XtNstring, buffer); i++;
    XtSetArg(Args[i], XtNeditType, XawtextEdit); i++;
    row = XtCreateManagedWidget("pinginterval", asciiTextWidgetClass, hostform
			      , Args, i);

    /* Add a "Dismiss" button */
    i = 0;
    XtSetArg(Args[i], XtNfromVert, row); i++;
    XtSetArg(Args[i], XtNfromHoriz, (Widget)NULL); i++;
    w = XtCreateManagedWidget("dismiss", commandWidgetClass, hostform
			      , Args, i);
    XtAddCallback(w, "callback", (XtCallbackProc)xt_remtopshell, (XtPointer)0);

    /* And finally a "Done" button */
    i = 0;
    XtSetArg(Args[i], XtNfromVert, row); i++;
    XtSetArg(Args[i], XtNfromHoriz, w); i++;
    w = XtCreateManagedWidget("do", commandWidgetClass, hostform
			      , Args, i);
    XtAddCallback(w, "callback", (XtCallbackProc)xt_addhost_do
		  , (XtPointer)hostform);

    XtSetMappedWhenManaged(shell, False);

    XtRealizeWidget(shell);

    /* Now fill in default netmask */
    addr.s_addr = DEFAULTNETMASK;
    sprintf(buffer, "%s", inet_ntoa(addr));
    i = 0;
    XtSetArg(Args[i], XtNstring, buffer), i++;
    XtSetValues(netmask, Args, i);

    XtPopup(shell, XtGrabNone);
}

XtCallbackProc xt_addinterface_do(w, hostinfostr)
Widget		w;
HOSTINFOSTR	*hostinfostr; {
    register int	i;
    char		*interface, *netmask;
    struct in_addr	hostaddr;
    u_long		hostnet;

    if(hostinfostr->his_hptr != (HOSTENTRY *)NULL
       && hostinfostr->his_topw != (Widget)NULL
       && inhostlist(hostinfostr->his_hptr)) {
	/* Gather all info */
	i = 0;
	XtSetArg(Args[i], XtNstring, &interface); i++;
	XtGetValues(XtNameToWidget(hostinfostr->his_topw, "interface")
		    , Args, i);
	
	i = 0;
	XtSetArg(Args[i], XtNstring, &netmask); i++;
	XtGetValues(XtNameToWidget(hostinfostr->his_topw, "netmask")
		    , Args, i);
	
	if(! *netmask)
	    hostnet = DEFAULTNETMASK;
	else if(index(netmask, '.'))
	    hostnet = inet_addr(netmask);
	else
	    sscanf(netmask, "%x", &hostnet);
	
	addifs(hostinfostr->his_hptr, interface, hostnet);
    }

    Flags |= PROG_KICKNET;
}

XtCallbackProc xt_addinterface_destroy(w, hostinfostr)
Widget		w;
HOSTINFOSTR	*hostinfostr; {

    xt_remtopshell(hostinfostr->his_topw);

    free(hostinfostr);
}

void xt_addinterface() {
    register int	i;
    char		buffer[80];
    Widget		form, shell, w;
    Widget		row;
    HOSTENTRY		*hptr;
    HOSTINFOSTR		*hostinfostr;
    struct in_addr	addr;

    if((hptr = widgettohost(Hostwidget)) == (HOSTENTRY *)NULL)
	return;

    if((hostinfostr = (HOSTINFOSTR *)malloc(sizeof(HOSTINFOSTR)))
       == (HOSTINFOSTR *)NULL) {
	sprintf(buffer, "Can't malloc %d bytes of memory! (addinterface)\n"
		, sizeof(HOSTINFOSTR));
	dispnotice("Add interface error", buffer);
	return;
    }

    sprintf(buffer, "%s", hptr->hl_name);

    i = 0;
    shell = XtCreatePopupShell(buffer, applicationShellWidgetClass
			       , Toplevel, Args, i);

    form = XtCreateManagedWidget("addinterfaceform", formWidgetClass, shell
				     , Args, i);

    hostinfostr->his_topw = form;
    hostinfostr->his_hptr = hptr;

    row = (Widget)NULL;

    /* Now we need to ask for hostname, netmask, community, pinginterval */
    sprintf(buffer, "Add interface to %s:", hptr->hl_name);

    i = 0;
    XtSetArg(Args[i], XtNfromVert, row); i++;
    XtSetArg(Args[i], XtNfromHoriz, row); i++;
    w = XtCreateManagedWidget("interface-prompt", labelWidgetClass, form
			      , Args, i);

    i = 0;
    XtSetArg(Args[i], XtNfromVert, row); i++;
    XtSetArg(Args[i], XtNfromHoriz, w); i++;
    XtSetArg(Args[i], XtNlabel, hptr->hl_name); i++;
    w = XtCreateManagedWidget("interface-hostname", labelWidgetClass, form
			      , Args, i);
    

    i = 0;
    XtSetArg(Args[i], XtNfromVert, row); i++;
    XtSetArg(Args[i], XtNfromHoriz, w); i++;
    XtSetArg(Args[i], XtNeditType, XawtextEdit); i++;
    row = XtCreateManagedWidget("interface", asciiTextWidgetClass, form
			      , Args, i);

    i = 0;
    XtSetArg(Args[i], XtNfromVert, row); i++;
    XtSetArg(Args[i], XtNfromHoriz, (Widget)NULL); i++;
    w = XtCreateManagedWidget("netmask-prompt", labelWidgetClass, form
			      , Args, i);

    addr.s_addr = DEFAULTNETMASK;
    sprintf(buffer, "%s", inet_ntoa(addr));
    i = 0;
    XtSetArg(Args[i], XtNfromVert, row); i++;
    XtSetArg(Args[i], XtNfromHoriz, w); i++;
    XtSetArg(Args[i], XtNstring, buffer); i++;
    XtSetArg(Args[i], XtNeditType, XawtextEdit); i++;
    row = XtCreateManagedWidget("netmask", asciiTextWidgetClass, form
			      , Args, i);

    /* Add a "Dismiss" button */
    i = 0;
    XtSetArg(Args[i], XtNfromVert, row); i++;
    XtSetArg(Args[i], XtNfromHoriz, (Widget)NULL); i++;
    w = XtCreateManagedWidget("dismiss", commandWidgetClass, form
			      , Args, i);
    XtAddCallback(w, "callback", (XtCallbackProc)xt_addinterface_destroy
		  , (XtPointer)hostinfostr);

    /* And finally a "Done" button */
    i = 0;
    XtSetArg(Args[i], XtNfromVert, row); i++;
    XtSetArg(Args[i], XtNfromHoriz, w); i++;
    w = XtCreateManagedWidget("do", commandWidgetClass, form
			      , Args, i);

    XtAddCallback(w, "callback", (XtCallbackProc)xt_addinterface_do
		  , (XtPointer)hostinfostr);

    XtPopup(shell, XtGrabNone);
}

void xt_version() {
    char	buf1[40], buf2[100];
    
    sprintf(buf1, "%s version", Myname);
    sprintf(buf2, "%s", Version);

    disptext(buf1, buf2);
}

struct Trafficstr {
    Widget		trs_widget;
    TRAFFICINFO		*trs_if;
    XtIntervalId	trs_timeid;
};

void xt_gettraffic(traffstr) struct Trafficstr *traffstr; {
    Widget	w;
    TRAFFICINFO	*traffif;

    if(Debug)
	printf("GetTraffic: getting info for interfaces:\n");

    for(traffif = traffstr->trs_if ; traffif ; traffif = traffif->tri_next) {
	if(traffif->tri_if->if_addr.s_addr == LOCALHOST) continue;
	if(traffif->tri_if->if_index == 127) continue; /* no index yet */
	if(Debug)
	    printf("%s: widget 0x%x\n", inet_ntoa(traffif->tri_if->if_addr)
		   , traffstr->trs_widget);
	snmp_gettraffic(traffif);
    }

    /* Call us again after a while */
    traffstr->trs_timeid = 
	XtAppAddTimeOut(XtWidgetToApplicationContext(Toplevel)
			, (u_long)(2 * 1000)
			, (XtTimerCallbackProc)xt_gettraffic
			, (XtPointer)traffstr);
}

void freetraffif(traffif) TRAFFICINFO *traffif; {

    if(! traffif) return;

    if(traffif->tri_next) freetraffif(traffif->tri_next);

    if(traffif->tri_inid)
	XtRemoveTimeOut(traffif->tri_inid);

    if(traffif->tri_outid)
	XtRemoveTimeOut(traffif->tri_outid);

    free(traffif);
}

XtCallbackProc xt_remtraffic(w, traffstr)
Widget			w;
struct Trafficstr	*traffstr; {

    if(traffstr->trs_timeid)
	XtRemoveTimeOut(traffstr->trs_timeid);

    freetraffif(traffstr->trs_if);

    free(traffstr);

    xt_remtopshell(w);
}

TRAFFICINFO *newtrafficinfo(ifptr, in_w, out_w)
IFENTRY		*ifptr;
Widget		in_w, out_w; {
    TRAFFICINFO	*traffif;
    
    traffif = (TRAFFICINFO *)malloc(sizeof(TRAFFICINFO));
    traffif->tri_if = ifptr;
    traffif->tri_inw = in_w;	/* place to write to later */
    traffif->tri_outw = out_w;	/* place to write to later */
    traffif->tri_next = (TRAFFICINFO *)NULL;
    traffif->tri_lastinval = 0;
    traffif->tri_lastintime = 0;
    traffif->tri_lastoutval = 0;
    traffif->tri_lastouttime = 0;
    traffif->tri_inid = 0;
    traffif->tri_outid = 0;

    return(traffif);
}

void xt_showtraffic() {
    register int	i;
    char		buffer[80];
    Widget		form, shell, w;
    Widget		row, col1, col2, col3;
    HOSTENTRY		*hptr;
    IFENTRY		*ifptr;
    struct Trafficstr	*traffstr;
    TRAFFICINFO		**atraffif;

    if((hptr = widgettohost(Hostwidget)) == (HOSTENTRY *)NULL)
	return;

    if((traffstr = (struct Trafficstr *)malloc(sizeof(struct Trafficstr)))
       == (struct Trafficstr *)NULL) {
	sprintf(buffer, "Can't malloc %d bytes of memory! (showtraffic)\n"
		, sizeof(struct Trafficstr));
	dispnotice("Showtraffic error", buffer);
	return;
    }

    i = 0;
    sprintf(buffer, "%s", hptr->hl_name);
    shell = XtCreatePopupShell(buffer, applicationShellWidgetClass
			       , Toplevel, Args, i);
    i = 0;
    form = XtCreateManagedWidget("trafficform", formWidgetClass, shell
				     , Args, i);

    traffstr->trs_widget = form; /* so we can find the display window */
    atraffif = &traffstr->trs_if; /* location for first if */

    row = (Widget)NULL;

    i = 0;
    sprintf(buffer, "%s", inet_ntoa(ifptr->if_addr));
    XtSetArg(Args[i], XtNfromVert, row); i++;
    XtSetArg(Args[i], XtNfromHoriz, (Widget)NULL); i++;
    col1 = XtCreateManagedWidget("Interface:      ", labelWidgetClass, form
			      , Args, i);

    i = 0;
    sprintf(buffer, "%s", inet_ntoa(ifptr->if_addr));
    XtSetArg(Args[i], XtNfromVert, row); i++;
    XtSetArg(Args[i], XtNfromHoriz, col1); i++;
    col2 = XtCreateManagedWidget("Packets In:         ", labelWidgetClass, form
			      , Args, i);

    i = 0;
    sprintf(buffer, "%s", inet_ntoa(ifptr->if_addr));
    XtSetArg(Args[i], XtNfromVert, row); i++;
    XtSetArg(Args[i], XtNfromHoriz, col2); i++;
    col3 = XtCreateManagedWidget("Packet Out:         ", labelWidgetClass, form
			      , Args, i);

    for(ifptr = hptr->hl_if ; ifptr ; ifptr = ifptr->if_next) {
	if(ifptr->if_addr.s_addr == LOCALHOST) continue;
	
	i = 0;
	sprintf(buffer, "%s", inet_ntoa(ifptr->if_addr));
	XtSetArg(Args[i], XtNfromVert, col1); i++;
	XtSetArg(Args[i], XtNfromHoriz, (Widget)NULL); i++;
	XtSetArg(Args[i], XtNlabel, buffer); i++;
	col1 = XtCreateManagedWidget("interfacename", labelWidgetClass, form
				  , Args, i);
	
	i = 0;
	XtSetArg(Args[i], XtNfromVert, col2); i++;
	XtSetArg(Args[i], XtNfromHoriz, col1); i++;
	XtSetArg(Args[i], XtNstring, "                    "); i++;
	col2 = XtCreateManagedWidget(buffer, asciiTextWidgetClass
				    , form, Args, i);

	i = 0;
	XtSetArg(Args[i], XtNfromVert, col3); i++;
	XtSetArg(Args[i], XtNfromHoriz, col2); i++;
	XtSetArg(Args[i], XtNstring, "                    "); i++;
	col3 = XtCreateManagedWidget(buffer, asciiTextWidgetClass
				    , form, Args, i);

	*atraffif = newtrafficinfo(ifptr, col2, col3);
	atraffif = &(*atraffif)->tri_next; /* place to put next one */
    }

    /* Add a "Dismiss" button */
    i = 0;
    XtSetArg(Args[i], XtNfromVert, col1); i++;
    XtSetArg(Args[i], XtNfromHoriz, (Widget)NULL); i++;
    w = XtCreateManagedWidget("dismiss", commandWidgetClass, form
			      , Args, i);

    /* Now add the routine to gather the information */
    traffstr->trs_timeid
	= XtAppAddTimeOut(XtWidgetToApplicationContext(Toplevel)
			  , (u_long)(2 * 1000)
			  , (XtTimerCallbackProc)xt_gettraffic
			  , (XtPointer)traffstr);
    
    XtAddCallback(w, "callback", (XtCallbackProc)xt_remtraffic
		  , (XtPointer)traffstr);

    XtPopup(shell, XtGrabNone);
}

XtCallbackProc xt_hostinfo_do(w, hostinfostr)
Widget		w;
HOSTINFOSTR	*hostinfostr; {
    register int	i;
    char		*name, *displayname, *interval, *community;
    HOSTENTRY		*hptr;
    Widget		widget;

    if((hptr = hostinfostr->his_hptr) != (HOSTENTRY *)NULL
       && (widget = hostinfostr->his_topw) != (Widget)NULL
       && inhostlist(hptr)) {
	/* Gather all info */
	i = 0;
	XtSetArg(Args[i], XtNstring, &name); i++;
	XtGetValues(XtNameToWidget(widget, "hostname"), Args, i);
	
	i = 0;
	XtSetArg(Args[i], XtNstring, &displayname); i++;
	XtGetValues(XtNameToWidget(widget, "displayname"), Args, i);
	
	i = 0;
	XtSetArg(Args[i], XtNstring, &interval); i++;
	XtGetValues(XtNameToWidget(widget, "interval"), Args, i);
	
	i = 0;
	XtSetArg(Args[i], XtNstring, &community); i++;
	XtGetValues(XtNameToWidget(widget, "community"), Args, i);
	
	/* See if any of it has changed */
	if(sstrcmp(displayname, hptr->hl_displayname) != 0) {
	    free(hptr->hl_displayname);
	    hptr->hl_displayname = cpystr(displayname);
/*	    XtSetArg(Args[0], XtNlabel, hptr->hl_displayname);
/*	    XtSetValues(hptr->hl_widget, Args, 1); */
	    XtDestroyWidget(hptr->hl_widget);
	    hptr->hl_widget = makehostwidget(displayname);
	    colorhostwidget(hptr, TRUE);
	    movehost(hptr, hptr->hl_x - hptr->hl_width / 2
		     , hptr->hl_y - hptr->hl_height / 2);
	}
	if(isdigit(*interval) && (i = atoi(interval)) != hptr->hl_interval)
	    hptr->hl_interval = i;
	if(sstrcmp(community, hptr->hl_community) != 0) {
	    free(hptr->hl_community);
	    hptr->hl_community = cpystr(community);
	}
    }

    Flags |= PROG_KICKNET;
}

XtCallbackProc xt_hostinfo_destroy(w, hostinfostr)
Widget		w;
HOSTINFOSTR	*hostinfostr; {

    xt_remtopshell(hostinfostr->his_topw);

    free(hostinfostr);
}

void xt_hostinfo() {
    register int	i;
    char		buffer[80];
    Widget		form, shell, w;
    Widget		row;
    HOSTENTRY		*hptr;
    HOSTINFOSTR		*hostinfostr;
    struct in_addr	addr;

    if((hptr = widgettohost(Hostwidget)) == (HOSTENTRY *)NULL)
	return;

    if((hostinfostr = (HOSTINFOSTR *)malloc(sizeof(HOSTINFOSTR)))
       == (HOSTINFOSTR *)NULL) {
	sprintf(buffer, "Can't malloc %d bytes of memory! (hostinfo)\n"
		, sizeof(HOSTINFOSTR));
	dispnotice("Host Info error", buffer);
	return;
    }

    sprintf(buffer, "%s Information", hptr->hl_name);

    i = 0;
    shell = XtCreatePopupShell(buffer, applicationShellWidgetClass
			       , Toplevel, Args, i);
    i = 0;
    form = XtCreateManagedWidget("hostinfoform", formWidgetClass, shell
				     , Args, i);

    hostinfostr->his_topw = form;
    hostinfostr->his_hptr = hptr;

    row = (Widget)NULL;

/*
 * Want:
 * Hostname:		name
 * Display:		name
 * Ping interval:	num
 * Community:		name
 * Last ping time:	time string
 */

    row = xt_promptline(form, "hostname-prompt", "hostname"
			, hptr->hl_name, row, XawtextRead);
    row = xt_promptline(form, "display-prompt", "displayname"
			, hptr->hl_displayname, row, XawtextEdit);
    sprintf(buffer, "%d", hptr->hl_interval);
    row = xt_promptline(form, "interval-prompt", "interval", buffer, row
			, XawtextEdit);
    row = xt_promptline(form, "community-prompt", "community"
			, hptr->hl_community, row, XawtextEdit);
    if(Debug)
	printf("current time is %d, lastpong is %d\n", time(0L)
	       , hptr->hl_lastpong);
    row = xt_promptline(form, "lastpong-prompt", "lastpong"
			, time2str(time(0L) - hptr->hl_lastpong), row
			, XawtextRead);

    /* Add a "Dismiss" button */
    i = 0;
    XtSetArg(Args[i], XtNfromVert, row); i++;
    XtSetArg(Args[i], XtNfromHoriz, (Widget)NULL); i++;
    w = XtCreateManagedWidget("dismiss", commandWidgetClass, form
			      , Args, i);
    XtAddCallback(w, "callback", (XtCallbackProc)xt_hostinfo_destroy
		  , (XtPointer)hostinfostr);

    /* And finally a "Do" button */
    i = 0;
    XtSetArg(Args[i], XtNfromVert, row); i++;
    XtSetArg(Args[i], XtNfromHoriz, w); i++;
    w = XtCreateManagedWidget("do", commandWidgetClass, form
			      , Args, i);

    XtAddCallback(w, "callback", (XtCallbackProc)xt_hostinfo_do
		  , (XtPointer)hostinfostr);

    XtPopup(shell, XtGrabNone);
}

makehostmenu() {
    register int	i, j;
    Widget		menu, widget;
    static struct {
	char	*menu_name;
	void	(*menu_routine)();
    }			menulist[]={
	{"Move host",		xt_menumove},
	{"DNS Hinfo",		xt_hinfo},
	{"Finger",		xt_finger},
	{"Show Interfaces",	xt_showinterfaces},
	{"Show pings",		xt_showpings},
	{"Delete this host",	xt_delhost},
	{"Add an interface",	xt_addinterface},
	{"Host Information",	xt_hostinfo},
/*	{"Showtraffic",		xt_showtraffic}, */
    };

    i = 0;
    menu = XtCreatePopupShell("hostmenu", simpleMenuWidgetClass, Toplevel
			      , Args, i);

    for(j = 0 ; j < ARRAYSIZE(menulist) ; j++) {
	i = 0;
	widget = XtCreateManagedWidget(menulist[j].menu_name
				       , smeBSBObjectClass, menu, Args, i);
	XtAddCallback(widget, "callback"
		      , (XtCallbackProc)menulist[j].menu_routine, 0);
    }
}

colorhostwidget(hptr, force) HOSTENTRY *hptr; int force; {
    u_long		nowtime;
    int			changed;
    
    time(&nowtime);
    changed = 0;
    if(nowtime - hptr->hl_lastpong > hptr->hl_interval * 4) {
	if(force || !(hptr->hl_flags & FL_DOWN)) {
	    if(Debug)
		printf("Host %s is down\n", hptr->hl_name);
	    hptr->hl_flags &= ~FL_UP;
	    hptr->hl_flags &= ~FL_MAYBE;
	    hptr->hl_flags |= FL_DOWN;
	    setblack(hptr);
	    changed++;
	}
    } else if(nowtime - hptr->hl_lastpong
	      > hptr->hl_interval * 2) {
	if(force || !(hptr->hl_flags & FL_MAYBE)) {
	    if(Debug)
		printf("Host %s is maybe down\n", hptr->hl_name);
	    hptr->hl_flags &= ~FL_UP;
	    hptr->hl_flags |= FL_MAYBE;
	    hptr->hl_flags &= ~FL_DOWN;
	    setgray(hptr);
	    changed++;
	}
    } else if(force || !(hptr->hl_flags & FL_UP)) {
	if(Debug)
	    printf("Host %s is back up\n", hptr->hl_name);
	hptr->hl_flags |= FL_UP;
	hptr->hl_flags &= ~FL_MAYBE;
	hptr->hl_flags &= ~FL_DOWN;
	setwhite(hptr);
	changed++;
    }

    return(changed);
}

x_readsocket() {
    static int		sequencenum=0;
    HOSTENTRY		*hptr;
    u_long		readping();
    u_long		nowtime;
    u_long		fromaddr;
    IFENTRY		*ifptr;
    int			i, j;
    int			changed;
    int			ret;
    char		buffer[128];
    PINGENTRY		*pingent;
	
    nowtime = time(0L);

    while((fromaddr = readping()) != (u_long)NULL) {
	if((hptr = iftohost(fromaddr)) != (HOSTENTRY *)NULL) {
	    /* this system's done */
	    if((pingent = extractping(fromaddr)) != (PINGENTRY *)NULL) {
		reinsertping(pingent, hptr->hl_interval);
		pingent->pen_pings = 0;	/* reset number of pings */
		hptr->hl_lastpong = nowtime;
	    }
	}
    }

    /* Set foregrounds, etc. based upon the results */
    changed = 0;
    for(hptr = Hostlist ; hptr ; hptr = hptr->hl_next)
	changed += colorhostwidget(hptr, FALSE);
    if(changed)
	XClearArea(Mydisplay, XtWindow(Form), 0, 0, Formwidth, Formheight
		   , True);

#ifdef	DEBUG
    printf("cleaning up children..."); fflush(stdout);
#endif
    /* Clean up old children */
    if(++sequencenum > 50) {
	while(wait3((union wait *)NULL, WNOHANG, (struct rusage *)NULL) > 0) ;
	sequencenum = 0;
    }

#ifdef	DEBUG
    printf("returning...\n");
#endif
    return(False);
}

makesnmpmenu() {
    register int	i, j;
    Widget		menu, widget;
    static struct {
	char	*menu_name;
	void	(*menu_routine)();
    }			menulist[]={
	{"sysDescr",		snmp_sysDescr},
	{"sysLocation",		snmp_sysLocation},
	{"sysContact",		snmp_sysContact},
	{"sysUptime",		snmp_sysUptime},
	{"ipAddrTable",		snmp_ipAddrTable},
	{"ipRouteTable",	snmp_ipRouteTable},
    };

    i = 0;
    menu = XtCreatePopupShell("snmpmenu", simpleMenuWidgetClass, Toplevel
			      , Args, i);

    for(j = 0 ; j < ARRAYSIZE(menulist) ; j++) {
	i = 0;
	widget = XtCreateManagedWidget(menulist[j].menu_name
				       , smeBSBObjectClass, menu, Args, i);
	XtAddCallback(widget, "callback"
		      , (XtCallbackProc)menulist[j].menu_routine
		      , (XtPointer)0);
    }
}
    
makemainmenu() {
    register int	i, j;
    Widget		menu, widget;
    static struct {
	char	*menu_name;
	void	(*menu_routine)();
    }			menulist[]={
	{"Exit",		xt_exitproc},
	{"Save configuration",	xt_save},
	{"Show All Pings",	xt_showallpings},
	{"Add host",		xt_addhost},
	{"Version",		xt_version},
    };

    i = 0;
    menu = XtCreatePopupShell("mainmenu", simpleMenuWidgetClass, Toplevel
			      , Args, i);

    for(j = 0 ; j < ARRAYSIZE(menulist) ; j++) {
	i = 0;
	widget = XtCreateManagedWidget(menulist[j].menu_name
				       , smeBSBObjectClass, menu, Args, i);
	XtAddCallback(widget, "callback"
		      , (XtCallbackProc)menulist[j].menu_routine
		      , (XtPointer)0);
    }
}
    
void xt_null() {}

void xt_canvasbutton(widget, event, params, num_params)
Widget		widget;
XButtonEvent	*event;
String		*params;
Cardinal	*num_params; {
    register int	i;
    Widget		w;
    NETENTRY		*netptr;
    Position		menu_x, menu_y;
    String		*parameters[3];

    /* See where the mouse was pushed */
    if(Debug)
	printf("canvas button event at %d, %d\n", event->x, event->y);

    /* Find the net which is there */
    if((netptr = findnet(event->x, event->y)) == (NETENTRY *)NULL)
	return;

    Netptr = netptr;

    w = XtNameToWidget(Toplevel, "netmenu");
    XtTranslateCoords(widget, event->x, event->y, &menu_x, &menu_y);
    i = 0;
    XtSetArg(Args[i], XtNx, menu_x), i++;
    XtSetArg(Args[i], XtNy, menu_y), i++;
    XtSetValues(w, Args, i);
 
    XtPopupSpringLoaded(w);
    XtGrabPointer(w, True, ButtonPressMask, GrabModeAsync, GrabModeAsync
		  , None, None, CurrentTime);

    return;
}

XtCallbackProc xt_netinfo_do(w, netinfostr)
Widget		w;
NETINFOSTR	*netinfostr; {
    register int	i;
    char		*name, *displayname, *speed, *type, *radius;
    int			nettype;
    Boolean		pinned;
    NETENTRY		*nptr;
    Widget		widget;

    if((nptr = netinfostr->nis_nptr) != (NETENTRY *)NULL
       && (widget = netinfostr->nis_topw) != (Widget)NULL
       && innetlist(nptr)) {
	/* Gather all info */
	i = 0;
	XtSetArg(Args[i], XtNstring, &name); i++;
	XtGetValues(XtNameToWidget(widget, "netname"), Args, i);
	
	i = 0;
	XtSetArg(Args[i], XtNstring, &speed); i++;
	XtGetValues(XtNameToWidget(widget, "netspeed"), Args, i);
	
/*	i = 0;
/*	XtSetArg(Args[i], XtNstring, &type); i++;
/*	XtGetValues(XtNameToWidget(widget, "nettype"), Args, i); */

	nettype = (int)XawToggleGetCurrent(XtNameToWidget(widget
							  , "linetoggle"));

	i = 0;
	XtSetArg(Args[i], XtNstate, &pinned); i++;
	XtGetValues(XtNameToWidget(widget, "pinnedtoggle"), Args, i);

	if(XtNameToWidget(widget, "netradius") != (Widget)NULL) {
	    i = 0;
	    XtSetArg(Args[i], XtNstring, &radius); i++;
	    XtGetValues(XtNameToWidget(widget, "netradius"), Args, i);

	}

	/* See if name has changed */
	if(sstrcmp(name, nptr->nl_name) != 0)
	    setnetname(nptr, name);

	/* check speed */
	if(isdigit(*speed) && (i = atoi(speed)) != nptr->nl_speed) {
	    if(i < 0) i = 0;
	    nptr->nl_speed = i;
	    if(i > 0)
		nptr->nl_flags |= NET_SPEEDSET;
	    else if(i == 0)
		nptr->nl_flags &= ~NET_SPEEDSET;
	}
	
	/* check type */
/*	lowerstring(type);	/* shift it to lower case */
/*	if(sstrcmp(type, nettypestring(nptr)) != 0) {
/*	    setnettype(nptr, type);
/*	} */

	if(nettype == TOGGLE_LINE) setnettype(nptr, "line");
	else setnettype(nptr, "circle");

	/* check radius if we have a circle */
	if((nptr->nl_flags & NET_CIRCLE) && isdigit(*radius)
	   && (i = atoi(radius)) != nptr->nl_radius) {
	    if(i < 0) i = 0;
	    nptr->nl_radius = i;
	}

	/* See if they want it "unpinned" */
	if(pinned == True && !(nptr->nl_flags & NET_PINNED))
	    nptr->nl_flags |= NET_PINNED;
	else if(pinned == False)
	    nptr->nl_flags &= ~NET_PINNED;
    }

    Flags |= PROG_KICKNET;
}

XtCallbackProc xt_netinfo_destroy(w, netinfostr)
Widget		w;
NETINFOSTR	*netinfostr; {

    xt_remtopshell(netinfostr->nis_topw);

    free(netinfostr);
}

void xt_menu_netinfo() {
    register int	i;
    char		buffer[80];
    Widget		form, shell, w;
    Widget		row;
    NETENTRY		*netptr;
    NETINFOSTR		*netinfostr;
    struct in_addr	addr;

    if(! Netptr) return;

    if((netinfostr = (NETINFOSTR *)malloc(sizeof(NETINFOSTR)))
       == (NETINFOSTR *)NULL) {
	sprintf(buffer, "Can't malloc %d bytes of memory! (netinfo)\n"
		, sizeof(NETINFOSTR));
	dispnotice("Net Info error", buffer);
	return;
    }

    if(Netptr->nl_name && *Netptr->nl_name)
	sprintf(buffer, "%s Network Information", Netptr->nl_name);
    else
	sprintf(buffer, "Network Information");

    i = 0;
    shell = XtCreatePopupShell(buffer, applicationShellWidgetClass
			       , Toplevel, Args, i);
    i = 0;
    form = XtCreateManagedWidget("netinfoform", formWidgetClass, shell
				     , Args, i);

    netinfostr->nis_topw = form;
    netinfostr->nis_nptr = Netptr;

    row = (Widget)NULL;

/*
 * Want:
 * Network Name:	name
 * Network number:	name
 * Network mask:	name
 * Network Speed:	number
 * Network Type:	typename
 * Ring radius:		number		(only if a circle)
 */

    row = xt_promptline(form, "netname-prompt", "netname"
			, Netptr->nl_name, row, XawtextEdit);
    row = xt_promptline(form, "netnumber-prompt", "netnumber"
			, inet_ntoa(Netptr->nl_net), row, XawtextRead);
    sprintf(buffer, "%d", Netptr->nl_speed);
    row = xt_promptline(form, "netspeed-prompt", "netspeed", buffer, row
			, XawtextEdit);
/*    row = xt_promptline(form, "nettype-prompt", "nettype"
/*			, nettypestring(Netptr), row, XawtextEdit); */
    sprintf(buffer, "%d", Netptr->nl_radius);
    row = xt_promptline(form, "netradius-prompt", "netradius"
			, buffer, row, XawtextEdit);

    /* Add toggle widgets (in a radio group) to line and circle types */
    i = 0;
    XtSetArg(Args[i], XtNfromVert, row), i++;
    XtSetArg(Args[i], XtNfromHoriz, (Widget)NULL), i++;
    w = XtCreateManagedWidget("nettype-prompt", toggleWidgetClass, form
				, Args, i);

    i = 0;
    XtSetArg(Args[i], XtNfromVert, row), i++;
    XtSetArg(Args[i], XtNfromHoriz, w), i++;
    XtSetArg(Args[i], XtNradioData, (XtPointer)TOGGLE_LINE), i++;
    if(! (Netptr->nl_flags & NET_CIRCLE))
	XtSetArg(Args[i], XtNstate, True), i++;
    w = XtCreateManagedWidget("linetoggle", toggleWidgetClass, form
				, Args, i);
    i = 0;
    XtSetArg(Args[i], XtNfromVert, row), i++;
    XtSetArg(Args[i], XtNfromHoriz, w), i++;
    XtSetArg(Args[i], XtNradioGroup, w), i++;
    XtSetArg(Args[i], XtNradioData, (XtPointer)TOGGLE_CIRCLE), i++;
    if(Netptr->nl_flags & NET_CIRCLE)
	XtSetArg(Args[i], XtNstate, True), i++;
    row = XtCreateManagedWidget("circletoggle", toggleWidgetClass, form
				, Args, i);
    
    i = 0;
    XtSetArg(Args[i], XtNfromVert, row), i++;
    XtSetArg(Args[i], XtNfromHoriz, (Widget)NULL), i++;
    if(Netptr->nl_flags & NET_PINNED)
	XtSetArg(Args[i], XtNstate, True), i++;
    row = XtCreateManagedWidget("pinnedtoggle", toggleWidgetClass, form
				, Args, i);

    /* Add a "Dismiss" button */
    i = 0;
    XtSetArg(Args[i], XtNfromVert, row); i++;
    XtSetArg(Args[i], XtNfromHoriz, (Widget)NULL); i++;
    w = XtCreateManagedWidget("dismiss", commandWidgetClass, form
			      , Args, i);
    XtAddCallback(w, "callback", (XtCallbackProc)xt_netinfo_destroy
		  , (XtPointer)netinfostr);

    /* And finally a "Do" button */
    i = 0;
    XtSetArg(Args[i], XtNfromVert, row); i++;
    XtSetArg(Args[i], XtNfromHoriz, w); i++;
    w = XtCreateManagedWidget("do", commandWidgetClass, form
			      , Args, i);

    XtAddCallback(w, "callback", (XtCallbackProc)xt_netinfo_do
		  , (XtPointer)netinfostr);

    XtPopup(shell, XtGrabNone);
}

void xt_netinfo(widget, event) Widget widget; XButtonEvent *event; {
    NETENTRY		*netptr;

    /* Find the net which is there */
    if((netptr = findnet(event->x, event->y)) == (NETENTRY *)NULL)
	return;

    Netptr = netptr;

    xt_menu_netinfo();
}

XtEventHandler xt_move_line_net(w, nptr, event)
Widget		w;
NETENTRY	*nptr;
XButtonEvent	*event; {
    
    if(Debug)
	printf("Set line network center to %d, %d\n", event->x, event->y);

    if(nptr->nl_starty == nptr->nl_endy) {
	/* horizontal net */
	nptr->nl_starty = event->y;
	nptr->nl_endy = event->y;
    } else {
	/* vertical net */
	nptr->nl_startx = event->x;
	nptr->nl_endx = event->x;
    }

    nptr->nl_flags |= NET_PINNED; /* we just set the location */

    XtRemoveEventHandler(Form, ButtonPressMask, False
			 , (XtEventHandler)xt_move_line_net
			 , (XtPointer)nptr);
    XtUngrabPointer(Form, CurrentTime);
    
    drawnets();
}

XtEventHandler xt_move_circle_net(w, nptr, event)
Widget		w;
NETENTRY	*nptr;
XButtonEvent	*event; {
    
    if(Debug)
	printf("Move circle network to %d, %d\n", event->x, event->y);

    nptr->nl_startx = event->x;
    nptr->nl_starty = event->y;
    nptr->nl_flags |= NET_PINNED; /* we forced the location */

    XtRemoveEventHandler(Form, ButtonPressMask, False
			 , (XtEventHandler)xt_move_circle_net
			 , (XtPointer)nptr);
/*    XtAddEventHandler(Form, ButtonPressMask, False
/*		      , (XtEventHandler)xt_save, (XtPointer)0); */
    XtUngrabPointer(Form, CurrentTime);

    drawnets();
}

void xt_menu_netlocation() {
    NETENTRY		*netptr;

    if(! Netptr) return;

    if(Netptr->nl_flags & NET_CIRCLE) {
	/* Case for a circle - locate new center */
	dispnotice("Reset location of circular network"
		   , "Press the mouse button where you want the new center");
	XtGrabPointer(Form, False, ButtonPressMask
		      , GrabModeAsync, GrabModeAsync
		      , XtWindow(Form), None, CurrentTime);
	
	XtRemoveEventHandler(Form, ButtonPressMask, False, xt_save, 0);
	XtAddEventHandler(Form, ButtonPressMask, False
			  , (XtEventHandler)xt_move_circle_net
			  , (XtPointer)Netptr);
    } else {
	/* Case for a line - choose horizontal or vertical */
	/* horizontal net */
	dispnotice("Reset location of line network"
		   , "Press the mouse button where you want the line");
	XtGrabPointer(Form, False, ButtonPressMask
		      , GrabModeAsync, GrabModeAsync
		      , XtWindow(Form), None, CurrentTime);
	XtRemoveEventHandler(Form, ButtonPressMask, False, xt_save, 0);
	XtAddEventHandler(Form, ButtonPressMask, False
			  , (XtEventHandler)xt_move_line_net
			  , (XtPointer)Netptr);
    }
}

void xt_netlocation(widget, event) Widget widget; XButtonEvent *event; {
    NETENTRY		*netptr;

    /* Find the net which is there */
    if((netptr = findnet(event->x, event->y)) == (NETENTRY *)NULL)
	return;

    Netptr = netptr;

    xt_menu_netlocation();
}

/*
 * Let's do some simple autodiscovery stuff.
 * (You can't do it this way -- way too slow!)
 */
/*void xt_menu_netdiscover() {
/*    register u_long	l;
/*    struct in_addr	addr;
/*    IFENTRY		*ifptr;
/*
/*    if(! Netptr) return;
/*
/*    printf("Autodiscovery! net=%s, ", inet_ntoa(Netptr->nl_net));
/*    printf("mask=%s\n", inet_ntoa(Netptr->nl_mask));
/*
/*    /* let's just insert some special ping list entries for all host numbers
/*       on this net which aren't already on the display.  We'll set a flag in
/*       the pinglist entry so that when/if it responds, it will be added to
/*       the display like normal, etc. */
/*
/*    for(l = 0 ; l < INFINITY ; l++) {
/*	if(l < 0) {
/*	    printf("l < 0\n");
/*	    break;
/*	}
/*	if((addr.s_addr = Netptr->nl_net.s_addr
/*	    | (~Netptr->nl_mask.s_addr & l))
/*	   == Netptr->nl_net.s_addr) continue;
/*	for(ifptr = Ifentries ; ifptr ; ifptr = ifptr->if_nextif)
/*	    if(ifptr->if_addr.s_addr == addr.s_addr) goto nextaddr; /* below */
/*	
/*	printf("Should ping IP address %s\n", inet_ntoa(addr));
/*	/* Create an ifentry */
/*	if((ifptr = newifentry(addr.s_addr, Netptr->nl_mask.s_addr))
/*	   == (IFENTRY *)NULL) {
/*	    dispnotice("Internal error!", "Can't malloc an ifentry!");
/*	    continue;
/*	}
/*	/* Set the special flag */
/*	ifptr->if_flags |= FL_DISCOVER;
/*    nextaddr: ;
/*    }
/*} */

makenetmenu() {
    register int	i, j;
    Widget		menu, widget;
    static struct {
	char	*menu_name;
	void	(*menu_routine)();
    }			menulist[]={
	{"Net information",	xt_menu_netinfo},
	{"Net location",	xt_menu_netlocation},
/*	{"Discover all hosts",	xt_menu_netdiscover}, */
    };

    i = 0;
    menu = XtCreatePopupShell("netmenu", simpleMenuWidgetClass, Toplevel
			      , Args, i);

    for(j = 0 ; j < ARRAYSIZE(menulist) ; j++) {
	i = 0;
	widget = XtCreateManagedWidget(menulist[j].menu_name
				       , smeBSBObjectClass, menu, Args, i);
	XtAddCallback(widget, "callback"
		      , (XtCallbackProc)menulist[j].menu_routine
		      , (XtPointer)0);
    }
}
    
/*
 * X action routines
 */
XtActionsRec	Actionlist[]={
    {"Selecthost",	xt_selecthost},
    {"Save",		xt_save},
    {"Movehost",	xt_movehost},
    {"Drawnets",	drawnets},
    {"Gridhost",	xt_gridhost},
    {"Finger",		xt_finger},
    {"Hinfo",		xt_hinfo},
    {"Exit",		xt_exitproc},
    {"Showif",		xt_showinterfaces},
    {"Addhost",		xt_addhost},
    {"Deletehost",	xt_delhost},
    {"Showallpings",	xt_showallpings},
    {"Addinterface",	xt_addinterface},
    {"Showpings",	xt_showpings},
    {"sysDescr",	snmp_sysDescr},
    {"ipAddrTable",	snmp_ipAddrTable},
    {"ipRouteTable",	snmp_ipRouteTable},
    {"sysLocation",	snmp_sysLocation},
    {"sysContact",	snmp_sysContact},
/*    {"Showtraffic",	xt_showtraffic}, */
    {"Version",		xt_version},
    {"Null",		xt_null},
    {"Netinfo",		xt_netinfo},
    {"Netlocation",	xt_netlocation},
    {"Canvasbutton",	xt_canvasbutton},
    {"Hostinfo",	xt_hostinfo},
};

xt_addactions() {

    XtAddActions(Actionlist, ARRAYSIZE(Actionlist));
}

/*
 * Flash this host widget to show we're sending a ping to it.
 */
flashhost(pent) PINGENTRY *pent; {
    char	*buffer;
    char	buf2[80];
    HOSTENTRY	*hptr;

    hptr = pent->pen_if->if_host;

    hptr->hl_flags &= ~FL_UP;
    hptr->hl_flags |= FL_MAYBE;
    hptr->hl_flags &= ~FL_DOWN;
    setgray(hptr);
}

void xt_resize(widget, w, event) Widget widget, w; XConfigureEvent *event; {
    register int	i;
    Position		width, height;
    Pixmap		newpixmap;

    if(event->type != ConfigureNotify) return;

    if(event->width == Formwidth && event->height == Formheight) return;

    /* Changed canvas size */
    if((newpixmap = XCreatePixmap(Mydisplay, XtWindow(Toplevel)
			  , event->width, event->height
			  , DefaultDepthOfScreen(Myscrptr))) == (Pixmap)NULL) {
	dispnotice("Error", "Can't enlarge pixmap for display!");
	return;
    }

    /* Make new pixmap all white */
    XSetForeground(Mydisplay, Gc, WhitePixelOfScreen(Myscrptr));
    XFillRectangle(Mydisplay, newpixmap, Gc, 0, 0
		   , event->width, event->height);

    /* Copy the old pixmap into the new one */
/*    XSetFunction(Mydisplay, Gc, GXcopy); */
    XCopyArea(Mydisplay, Lines, newpixmap, Gc, 0, 0, Formwidth, Formheight
	      , 0, 0);

    /* Free old one, reassign new values */
    XFreePixmap(Mydisplay, Lines);
    Lines = newpixmap;
    Formwidth = event->width;
    Formheight = event->height;

    i = 0;
    XtSetArg(Args[i], XtNbackgroundPixmap, Lines); i++;
    XtSetValues(Form, Args, i);

    /* Because this is what's expected elsewhere */
/*    XSetFunction(Mydisplay, Gc, GXset); */

    XClearArea(Mydisplay, XtWindow(Form), 0, 0, Formwidth, Formheight, True);
    
    return;
}
