#include <stdio.h>
#include <ctype.h>
#include <pwd.h>
#include <sys/signal.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/nameser.h>
#include <resolv.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/socket.h>

#include <X11/Xlib.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Dialog.h>
#include <X11/Xaw/Paned.h>
#include <X11/Xaw/Grip.h>
#include <X11/Xaw/Viewport.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Toggle.h>
#include <X11/Xaw/Scrollbar.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/SmeBSB.h>
#include <X11/Xaw/MenuButton.h>
#include <X11/Shell.h>
#include <X11/bitmaps/gray>

#undef	TRUE
#undef	FALSE
#include "smp.h"

int	xt_ping();		/* ping.c */

/*
 * Global declarations
 */
#define DEFAULTNETMASK	htonl(0xffffff00)
#define DEFAULTPINGTIME	5
#define DEFAULTCOMMUNITY	"public"
#define PINGBACKOFF	100	/* if > MAXPINGS, try every PINGBACKOFF's */
#define MAXPINGS	100	/* max. pings before we give up for a while */
#define NETSPACING	4
#define INFINITY	((int)(((unsigned)(-1)) >> 1))
#define SMALL_INFINITY	((int)(((u_short)(-1)) >> 1))
#define SLACKOFF	4
#define TEXTALLOCINC	128
#define LOCATIONS	".xsnmp_locations" /* where host loc.'s are stored */
#define HOSTLIST	".xsnmp_hosts" /* where hostnames are stored */
#define XSNMPRC		".xsnmprc" /* new form for the above files */
#define LOCALHOST	htonl(0x7f000001) /* 127.0.0.1 */
#define SNMPQUERYINTERVAL	(60) /* in seconds */
#define SNMPQUERYTIMEOUT	(15) /* in seconds */
#define SNMPMAXTRIES	(5)	/* max. times to try an ipAddrTable query */
#define SNMPTRYAGAIN	(60 * 60) /* try SNMP again every hour */
#define LEGENDX		70	/* X location of legend (upper left corner) */
#define LEGENDY		20	/* Y location of legend start (upper left) */
#define LEGENDLINEHEIGHT	15 /* distance between lines */
#define SCROLLBARWIDTH	15	/* actually it's 14 by default */
#define NOTICEDISPTIME	5	/* sec's to display a notice */
#define INITIAL_X	200	/* inital x for new hosts */
#define INITIAL_Y	0	/* inital y for new hosts */
#define NETSELECTDIST	3	/* allowed dist. from net for selection */
#define DEFAULTRADIUS	10	/* default radius for a circle net */
#define TESTIFUPTIME	(60*60*1) /* check every hour for interfaces which
				   may have come back up */

#define TOGGLE_LINE	1
#define TOGGLE_CIRCLE	2

#define FL_PONG		(1 << 0) /* We got a ping reply from this guy */
#define FL_UP		(1 << 1) /* This host is up */
#define FL_MAYBE	(1 << 2) /* May be up, didn't respond last time */
#define FL_DOWN		(1 << 3) /* This host is down */
#define FL_PINGNEXT	(1 << 4) /* This host MUST be pinged next */
#define FL_SNMP		(1 << 5) /* Interface is SNMP derived */
#define FL_SNMPDONE	(1 << 6) /* SNMP ipAddrTable completed */
#define FL_NOSNMP	(1 << 7) /* this host doesn't understand SNMP */
#define FL_SHOWPINGS	(1 << 8) /* show pings for this host */
#define FL_INCLUDE	(1 << 9) /* came from an included 'rc file */
#define FL_SNMPMASK	(1 << 10)/* netmask came from snmp */
#define FL_CONFMASK	(1 << 11)/* netmask came from configuration info */
#define FL_DISCOVER	(1 << 12)/* used during autodiscovery */

/* Network flags */
#define NET_PINNED	(1 << 0) /* This network line is pinned down */
#define	NET_INCLUDE	(1 << 1) /* came from an include 'rc file */
#define NET_SPEEDSET	(1 << 2) /* speed set from .xsnmprc file */
#define NET_CIRCLE	(1 << 3) /* network should be a circle */
#define NET_SETSTART	(1 << 4) /* we have set the start x,y only */
#define NET_LINE	(1 << 5) /* network should be a line */

/* Program flags */
#define PROG_KICKNET	(1 << 0) /* redraw network - new net connection */
#define PROG_DEBUG	(1 << 1) /* debug command line option */
#define PROG_COLOR	(1 << 2) /* this display has color */
#define PROG_DEBUG_X	(1 << 3) /* debug the X routines */
#define PROG_DEBUG_SNMP	(1 << 4) /* debug the snmp routines */
#define PROG_DEBUG_IF	(1 << 5) /* debug the if routines */
#define PROG_DEBUG_NET	(1 << 6) /* debug the net routines */
#define PROG_DEBUG_HOST	(1 << 7) /* debug the host routines */
#define PROG_DEBUG_DNS	(1 << 8) /* debug the dns routines */
#define PROG_DEBUG_FINGER (1 << 9) /* debug the finger routines */
#define PROG_DEBUG_PING (1 << 10) /* debug the ping routines */
#define PROG_DEBUG_MAIN (1 << 11) /* debug the main routines */
#define PROG_DEBUG_YACC	(1 << 12) /* debug the yacc parsing routines */
#define PROG_WINDOW	(1 << 13) /* we have a window in the screen now */
#define PROG_REVERSE	(1 << 14) /* reverse video */

typedef struct Addrstr {
    struct in_addr	if_addr;
    struct in_addr	if_net;
    struct in_addr	if_mask;
    u_long		if_lastsnmp;
    struct Addrstr	*if_nextif; /* pointing to next IF */
    struct Addrstr	*if_next; /* pointing to next IF for this host */
    int			if_flags;
    struct Hostentry	*if_host; /* hostentry for this interface */
    int			if_speed; /* speed of this interface */
    char		if_index; /* SNMP interface index */
    u_long		if_lastup; /* time we discovered this was down */
}		IFENTRY;
IFENTRY			*Ifentries;

typedef struct Hostentry	{
    char		*hl_name;
    char		*hl_displayname;
/*    struct sockaddr_in	hl_addr; */
/*    u_long		hl_net; */
    IFENTRY		*hl_if;
    int			hl_interval;
    u_long		hl_lastpong;
    int			hl_flags;
    Position		hl_x;
    Position		hl_y;
    Dimension		hl_width;
    Dimension		hl_height;
    Widget		hl_widget;
    char		*hl_community;
    u_long		hl_lastsnmpif;
    u_long		hl_lastsnmp;
    struct Hostentry	*hl_next;
}		HOSTENTRY;
extern HOSTENTRY	*Hostlist;

typedef struct Pingentry {
    IFENTRY		*pen_if;
    struct Pingentry	*pen_next;
    int			pen_interval;
    u_long		pen_pings;
    int			pen_flags;
}		PINGENTRY;
extern PINGENTRY	*Pinglist;

int			Pingsocket;
    
typedef struct Netliststr	{
    struct Netliststr	*nl_next;
    char		*nl_name;
    struct in_addr	nl_net;
    struct in_addr	nl_mask;
    int			nl_speed; /* speed from SNMP info. */
    short		nl_minx;
    short		nl_miny;
    short		nl_maxx;
    short		nl_maxy;
    short		nl_startx;
    short		nl_starty;
    short		nl_endx;
    short		nl_endy;
    int			nl_flags;
    short		nl_radius;
}		NETENTRY;
extern NETENTRY		*Netlist;
int	Lastnet;

struct Coord {
    short	x;
    short	y;
};

typedef struct Trafficinfo {
    IFENTRY		*tri_if;
    Widget		tri_inw;
    Widget		tri_outw;
    struct Trafficinfo	*tri_next;
    float		tri_lastinval;
    u_long		tri_lastintime;
    float		tri_lastoutval;
    u_long		tri_lastouttime;
    int			tri_flags;
    XtIntervalId	tri_inid;
    XtIntervalId	tri_outid;
}		TRAFFICINFO;

/* Storage location for snmp query results */
typedef struct Snmpresult	{
    SmpRequestType	*smr_req;
    struct Snmpresult	*smr_next;
}		SNMPRESULT;

typedef struct Snmpidstr {
    struct Snmpidstr	*snmp_next;
    XtInputId		snmp_id;
    XtIntervalId	snmp_timeid;
    HOSTENTRY		*snmp_host;
    SmpBindType		*snmp_bind;
    SmpNameType		*snmp_orig;
    SmpLengthType	snmp_origlen;
    SmpRequestType	*snmp_req;
    short		snmp_count;
    short		snmp_flags;
    int			snmp_socket;
    SmpSocketType	snmp_udp;
    char		*snmp_buf;
    char		*snmp_glue;
}		SNMPIDSTR;

/* Flag definitions */
#define SNMP_DONE	(1 << 0) /* SNMP query is done now */
#define SNMP_GLUE	(1 << 1) /* make sure you free this */

typedef struct {
    int		sc_speed;
    char	*sc_name;
    XColor	sc_color;
}		SPEEDCOLOR;

typedef struct {
    int		lt_speed;
    int		lt_style;
    int		lt_width;
}		LINETYPE;

typedef struct {
    struct in_addr	rts_lastroute;
    char		*rts_buf;
    char		*rts_end;
    int			rts_size;
}		ROUTETXTSTR;

typedef struct Hostinfostr {
    Widget	his_topw;
    HOSTENTRY	*his_hptr;
}		HOSTINFOSTR;

typedef struct Netinfostr {
    Widget	nis_topw;
    NETENTRY	*nis_nptr;
}		NETINFOSTR;

typedef struct Parsehost {
    char	*pah_name;	/* host "name" */
    char	*pah_display;	/* display "name" */
    u_long	pah_mask;	/* mask <num> (hex or decimal) */
    int		pah_interval;	/* interval <num> */
    char	*pah_community;	/* community "string" */
    short	pah_x;		/* loc x y */
    short	pah_y;		/* " */
    int		pah_flags;	/* random flags */
}		PARSEHOST;

typedef struct Parsenet {
    u_long	pan_addr;	/* network address */
    char	*pan_name;	/* network "name" */
    short	pan_startx;	/* startloc x y */
    short	pan_starty;	/* " */
    short	pan_endx;	/* endloc x y */
    short	pan_endy;	/* " */
    int		pan_speed;	/* speed <num> */
    int		pan_radius;	/* radius <num> */
    short	pan_centerx;	/* center x y */
    short	pan_centery;	/* " */
    int		pan_flags;	/* random flags */
}		PARSENET;

typedef struct Parseif {
    char	*pai_name;	/* interface "name" */
    u_long	pai_addr;	/* ipaddr <num> (hex or decimal) */
    u_long	pai_mask;	/* mask <num> (hex or decimal) */
    int		pai_flags;	/* random flags */
}		PARSEIF;

/* Globally available resource information */
typedef struct {
    char	*geometry;
}	RESOURCES;

/* Global routines and where they're defined */
void		xt_exitproc();		/* x.c */
HOSTENTRY	*iftohost();		/* host.c */
NETENTRY	*iftonet();		/* host.c */
HOSTENTRY	*widgettohost(); 	/* host.c */
Widget		readhosts();		/* host.c */
char		*fullhostname(); 	/* host.c */
HOSTENTRY	*addhost();		/* host.c */
HOSTENTRY	*nametohost();		/* host.c */
PARSEHOST	*newparsehost();	/* host.c */
Widget		makehostwidget();	/* host.c */
char		*appendstr();		/* util.c */
int		x_readsocket();		/* x.c */
void		drawnets();		/* net.c */
PINGENTRY	*extractping();		/* ping.c */
Widget		disptext();		/* x.c */
void		xt_hinfo();		/* dns.c */
void		xt_finger();		/* finger.c */
int		xt_snmp();		/* snmp.c */
XtInputCallbackProc x_readsnmp();	/* snmp.c */
char		*skipspaces();		/* util.c */
char		*cpystr();		/* util.c */
char		*time2str();		/* util.c */
IFENTRY		*addif();		/* if.c */
IFENTRY		*addifmask();		/* if.c */
IFENTRY		*addifindex();		/* if.c */
IFENTRY		*addifspeed();		/* if.c */
IFENTRY		*ifindextoif();		/* if.c */
PARSEIF		*newparseif();		/* if.c */
IFENTRY		*newifentry();		/* if.c */
void		xt_resize();		/* x.c */
NETENTRY	*findnet();		/* net.c */
NETENTRY	*newnetentry();		/* net.c */
PARSENET	*newparsenet();		/* net.c */
char		*nettypestring(); 	/* net.c */

/* SNMP routines */
void		snmp_sysDescr(); 	/* snmp.c */
void		snmp_sysLocation(); 	/* snmp.c */
void		snmp_sysContact(); 	/* snmp.c */
void		snmp_sysUptime();	/* snmp.c */
void		snmp_ipAddrTable(); 	/* snmp.c */
void		snmp_ipRouteTable(); 	/* snmp.c */

/* Global variables */
Widget		Toplevel, Form;		/* x.c */
Widget		Exit;			/* x.c */
extern Arg	Args[];			/* x.c */
Colormap	Cmap;			/* x.c */
XColor		White, Black, Gray;	/* x.c */
Display		*Mydisplay;		/* x.c */
int		Myscreen;		/* x.c */
Screen		*Myscrptr;		/* x.c */
Pixmap		Lines;			/* x.c */
GC		Gc;			/* x.c */
Widget		Hostwidget;		/* x.c */
int		Flags;			/* main.c */
char		*Myname;		/* main.c */
char		*Version;		/* main.c */
Dimension	Formwidth;		/* x.c */
Dimension	Formheight;		/* x.c */
RESOURCES	Resources;		/* main.c */
extern int	yylineno;		/* in yacc routines */
FILE		*Xsnmprcfile;		/* main.c */
char		*Xsnmprcp;		/* main.c */
char		*Xsnmprcname;		/* main.c */

#define ARRAYSIZE(a)	(sizeof(a) / sizeof(*a))
#ifndef	MAX
#  define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#endif	/* MAX */

#ifndef	MIN
#  define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#endif	/* MIN */
