.lf 1 xsnmp.m
.TH XSNMP
.SH NAME
xsnmp - display a network and do SNMP queries
.SH SYNOPSIS
.B "xsnmp"
[-option ...] [-D?]
.SH DESCRIPTION
Xsnmp is a simple network monitoring program.  Its primary purpose is to
graphically display various networks (including routers and various subnets)
and show information about systems on those networks.  Primarily it uses the
ICMP "echo" protocol to "ping" each system at regular intervals (which can be
individually specified per system).  If a system hasn't responded to a "ping"
after double the ping interval, it is marked as "possibly down".  If it still
hasn't responded after 4 times the ping interval, it is marked as "down".
Systems which are "possibly down" are colored grey and systems which are marked
as "down" are colored black.  All others are colored white.

In addition to the "pinging" described above, the program also uses the SNMP
protocol to query systems as to their interfaces.  By this method it can
determine what systems are connected to what subnets as well as the speeds of
the networks.  Network lines are automatically drawn between systems to show
their network association and the colors are used to represent network speeds.
(On monochrome systems, differing types of lines are used instead of colors.)

There are various other functions contained in the program as well which may
be useful for system administration.  For example, there is a "finger"
function which will do the command "finger @system" and place the output into
a window which will pop up when the finger is completed.  There also is
a "hostinfo" function which will do a DNS "HINFO" query on the system in order
to determine what type of system it is and what OS it is running.

Other functions will be added as time permits.  For example, popping up a
window to monitor network traffic through a specific interface on a system.
SNMP is currently used to gather and display the routing table for a host.
Eventually it would be useful to be able to \fIset\fP routing entries as well.

The program used to expect to read a \fI~/.xsnmp_hosts\fP file which contained
needed information about the hosts it should display and monitor.  It also
expected to read a \fI~/.xsnmp_locations\fP file containing the X and Y
coordinates of the hosts on the display.

All versions of the program starting with version 2.2 will expect to read
a \fI~/.xsnmprc\fP file which contains the configuration information.
The end user should never have to actually edit this file directly.  The
"Save()" function will write the file for you.  For backward compatibility
the program will look for the \fI~/.xsnmp_hosts\fP and \fI~/.xsnmp_locations\fP
files if there is no \fI~/.xsnmprc\fP file.  For further information about the
form of the \fI~/.xsnmprc\fP file, see the section
entitled ".xsnmprc File Syntax" below.

Built in X Functions:
.TP 8
.B Sethostname()
Used before other functions to specify to which host the next function should
be applied.  For example, when a mouse button is pressed while positioned
on a host on the display, the "Sethostname" function specifies that the
hostname associated with the host currently under the cursor is the one to
which the next function invoked should be applied.  See the Xsnmp app-defaults
file for a complete example.
.TP 8
.B Save()
Saves the current network configuration.  Writes a ~/.xsnmp_hosts and
a ~/.xsnmp_locations file.  All hosts currently displayed will be written to
both files.  Additional network interfaces which were specified by the user
will be written to ~/.xsnmp_hosts but interfaces gathered from SNMP information
will not be written out (since they should be gathered dynamically).
.TP 8
.B Movehost()
Allows hosts on the display to be moved using the mouse.  Invoking this
function will make the currently set host (see "Sethostname") follow the
cursor.  It is intended to be called from a button motion event.  See the
app-defaults file for specific use.
.TP 8
.B Drawnets()
Invoking this function redraw the entire network display.  It is intended to
be used after functions such as "Movehost" in order to redraw the network
following the move.
.TP 8
.B Gridhost()
This function will move the currently selected host (see "Sethostname") such
that its position is a multiple of 10 pixel units.  This is helpful in aligning
hosts on the display.  It should be called on the Button "Up" event after
the Movehost function has been done.  See the app-defaults file for an
example.
.TP 8
.B Finger()
This function will run the command "finger @system" where "system" is the
fully qualified hostname of the currently selected host (see "Sethostname").
The output from this command will be placed into a window which will be popped
up on the screen.
.TP 8
.B Hostinfo()
This function will do a Domain Nameserver System (DNS) "HINFO" query on the
currently selected host (see "Sethostname") and display the output in a popup
window on the screen.
.TP 8
.B Exit()
Exit the program.  This does \fInot\fP save the configuration.  In order to
save the configuration you must use the "Save" function first.
.TP 8
.B Showif()
Shows all network interfaces associated with the currently selected host
(see "Sethostname").  The popup window contains the network IP address, network
mask, interface number (used in some SNMP queries), and interface speed.  Also
there is a designation of whether the interface was "hardcoded" (specified in
the ~/.xsnmp_hosts file) or was gathered from SNMP queries.
.TP 8
.B Addhost()
Pops up a dialog window which can be used to add hosts to the display.  All
hosts added will first appear at the top of the display and can then be moved
to the appropriate place with the Movehost function.  If any parts of the
dialog window are not filled in, they will be assumed to be the compiled in
defaults.  Netmask can be specified in either hex or decimal notation.
.TP 8
.B Deletehost()
Removes the currently selected host (set "Sethostname") from the display also
removing all information about the host's interfaces, etc.
.TP 8
.B Showallpings()
Toggles the "show all pings" flag for all hosts on the screen.  See
the "Showpings" function.
.TP 8
.B Addinterface()
Adds an interface description to the currently selected
host (see "Sethostname").  The only information which need be supplied is
the IP address.  If the network mask is not supplied, it will be set to the
compiled in default value.  If the network mask is supplied, it can be
specified in either hex or decimal notation.
.TP 8
.B Showpings()
Toggle the "show pings" flags for the currently selected
host (see "Sethostname").  When this flag is set, every time an ICMP echo
packet is sent to this host, the displayed box will change to grey.  When the
ICMP echo reply comes back, the host will turn back to white.  This usually
happens rather quickly, so what the user sees is a small flash of the
appropriate hostname on the display.  NOTE, this is a \fItoggle\fP so it can
be turned back off again by simply envoking this function again.  Note also
that the "Showallpings" function (above) simply executes the "Showpings"
function for all hosts currently on the display.  This means that if you have
done "Showpings" for one host and then do "Showallpings", the "show pings"
flag for the first host will now be turned off.  (This may change in the
future since it seems counter-intuitive to some people.)
.TP 8
.B Version()
Display version information about the currently running xsnmp program.
.TP 8
.B Null()
This routine does nothing.  It is bound to the Return key on various popup
forms so that pressing the Return key will not do anything.  If this binding
isn't made, the user may think that pressing Return has somehow done some
action, but instead it simply scrolls the information past the top of the
window.  This is very confusing.  Thus, the need for the Null() routine.
.TP 8
.B Netinfo()
Popup a form containing information about the currently selected network.
This form allows the user to edit various types of information about the
network and apply those changes.  See "Netinfo Form" below.
.TP 8
.B Canvasbutton()
This routine is used internally and probably has no reason to be used
elsewhere.  It consists of a simple routine which checks to see if the mouse
is currently positioned on or near a network line on the screen.  If it is,
a pointer to that network information is saved and
the menu named "netmenu" is popped up and selections are allowed from it.
.sp
.TP 8
\fISNMP functions\fP
.TP 8
.B sysDescr()
Do an SNMP "sysDescr" query and display the results.  This is supposed to be
a "system description" and usually contains information about the manufacturer
of the system and what OS and version it's running.
.TP 8
.B ipAddrTable()
Do an SNMP "ipAddrTable query.
This tells the program to query the currently selected host for its network
interface table.  This is an "aggregate" query and thus requires
the use of a ("powerful") getNext operator.  Thus, multiple queries must be
made and any one of them may timeout.  All interface information retrieved is
added to the interface information for that host.  Once a new IP address is
added in this fashion, the program will automatically proceed to send out
SNMP queries to retrieve the network mask, network interface number, network
speed, etc.  The ipAddrTable query is automatically done by the program and
thus need never be actually invoked by the user unless it is desired to simply
speed up the process.
.TP 8
.B ipRouteTable()
Do an SNMP "ipRouteTable" query and display the results.  This retrieves the
system's routing table via multiple ("powerful") getNext operatations.  If one
of the queries times out, it will be repeated and a message will be displayed
stating such.  When the entire table has been retrieved it will be displayed
in a popup window.  If the queries timeout repeatedly such that the entire
table cannot be retrieved, the partial list of already gathered information
will be displayed with a note stating such.
.SH DEFAULT MENUS AND COMMAND BINDINGS
The Xsnmp app-defaults file contains all of the bindings between the builtin
commands and the mouse buttons and the keyboard.  This describes the default
bindings.
.TP 8
.B sysLocation()
Do an SNMP "sysLocation" query and display the results.  This query should
return the physical location of the system, however the system administrator
needs to set this information in the SNMP server in order for there to be
anything useful here.
.bp
.TP 8
.B sysContact()
Do an SNMP "sysContact" query and display the result.  This query should
return information about who to contact w.r.t. managing this system; for
example, if there are problems with it, etc.  This information needs to be
set by the system adminstrator in order for there to be anything useful here.

The following discussion of how mouse and keyboard buttons work is assuming
that the program is using the default app-defaults information as it is
distributed.  Your local setup may differ.

.SH Mouse Button Bindings

There is a button in the upper left of the display window which should be
labeled "Functions" and when pressed will display a short menu:

.nf
Menu item                             Function
-------------------------------------------------------------
Exit                                  Exit() function
Save Configuration                    Save() function
Show All Pings                        Showallpings() function
Add Host                              Addhost() function
Version                               Version() function
.fi

If the mouse is currently located on one the hostname boxes, the following
menu will be displayed when the left button (button 1) is pressed:

.nf
Menu item                             Function
-------------------------------------------------------------
Move host                             Movehost()
DNS Info                              Hinfo()
Finger                                Finger()
Show Interfaces                       Showif()
Show Pings                            Showpings()
Delete this host                      Deletehost()
Add an interface                      Addinterface()
Host Information                      Hostinfo()
.fi

.bp
If the mouse is currently located on one the hostname boxes, the following
SNMP functions are made available when the middle button (button 2) is pressed:

.nf
Menu item                             Function
-------------------------------------------------------------
sysDescr                              sysDescr()
sysLocation                           sysLocation()
sysContact                            sysContact()
sysUptime                             sysUptime()
ipAddrTable                           ipAddrTable()
ipRouteTable                          ipRouteTable()
.fi

If the mouse is currently located on the backgound on top of a network line
or on a network circle, the following menu will be made available when the
left button (button 1) is pressed:

.nf
Menu item                             Function
-------------------------------------------------------------
Net Information                       Netinfo()
Net Location                          Netlocation()
.fi

If the mouse is currently located on one the hostname boxes, the
right button (button 3) may be used to "drag and drop" the host to a different
position on the display.

.SH Key Bindings

In additional to the above, the following functions are bound to keys when the
mouse is positioned on the background:

.nf
Key                                   Function
--------------------------------------------------------
S                                     Save()
A                                     Addhost()
P                                     Showallpings()
E                                     Exit()
V                                     Version()
.fi

.bp
The following functions are bound to keys when the mouse is positioned on a
hostname box:

.nf
Key                                   Function
--------------------------------------------------------
f                                     Finger()
h                                     Hostinfo()
D                                     Deletehost()
i                                     Showif()
a                                     Addinterface()
c                                     sysContact()
l                                     sysLocation()
.fi
.SH Popup Forms
Various windows may be "popped up" during the running of the program depending
upon what actions are taken by the user and what functions are invoked.  The
following is a discussion of these popup forms:

.B Disptext
Any function which pops up a window full of text is using the "disptext"
subroutine.  These windows will always have a "Dismiss" button at the bottom
which, when pressed, will remove the window from the screen.  In addition,
if the text window would have appeared larger than 2/3 the size of the X
server's screen, a scroll bar is automatically added and the window is limited
to that size.

.B Dispnotice
Anytime the program needs to tell the user something for which there is no
reason it should remain on the screen more than a little while (such as an
error message), the "dispnotice" routine is used.  This routine functions the
same as the "disptext" routine described above but there will be no "Dismiss"
button.  Instead the entire window will simply "timeout" and be removed from
the screen automatically.  The time it remains on the screen is a compile-time
variable.  (At some point in the future it will probably be setable from
the command line or .Xdefaults file.)

.B Addhost
The "addhost" form is used by the function "Addhost()" in order to collect
the needed information for adding a new host box onto the screen.
All defaults will be automatically filled in when the form pops up.
The fields are:

.TP 8
.B Hostname
\- Either a hostname or IP address.  If a hostname is given, it is
looked up with the "gethostbyname" routine in order to get the IP address of
that host.  If an IP address is given, it is used directly.

.TP 8
.B Display Name
\- The name which should be used on the display for this host.

.TP 8
.B Net Mask
\- The subnet mask for this host.  It may be given in either hex
notation if preceeded with "0x" or decimal notation.  For example, the
specifications "0xffffff00" and "255.255.255.0" refer to the same subnet
mask value.

.TP 8
.B Community
\- The SNMP community name for this host.  Only "read" SNMP queries
are done, so usually a community name of "public" or "monitor" will work.  If
you are unsure of the community name for a host and these do not work, you
should contact the system administrator for that system or the network
administrator for your network.

.TP 8
.B Ping Interval
\- The number of seconds between sending an ICMP Echo request
to this host.  If the host is relatively important, it should be "pinged" more
often.  If not, it should be "pinged" less often.

.PP
.B Addinterface
The "addinterface" form is used by the function "Addinterface()" in order
to collect
the needed information for adding an additional interface to the currently
selected host.  This may be needed if either the host doesn't support SNMP
(isn't running an SNMP server for example; Sun workstations don't seem to
have one), or if the SNMP community name is incorrect or not known.
All defaults will be automatically filled in when the form pops up.
The fields are:

.TP 8
.B Add interface to ...
\- Fill in either the IP address or a name which will
map to the IP address of the additional interface for this host.

.TP 8
.B Subnet mask
\- This should be filled in as the subnet mask for the additional
interface.  Just as described above w.r.t. the addhost form, the subnet mask
may be specified in either hex or decimal notation.

.PP
.B Hostinfo
The "hostinfo" form is used by the function "Hostinfo()" in order to both
display host information and allow the user to make changes to it.
The fields are:

.TP 8
.B Hostname
\- This is the full official hostname (or IP address) of the host.  It
cannot be changed because to do so would refer to a totally different host!
If the desire is to change the hostname which is displayed on the screen, see
below.

.TP 8
.B Display name
\- This is the name which is currently being used on the display
for this host.  If you change this name, the name on the screen will change
accordingly.

.TP 8
.B Ping Interval
\- This is the "ping" interval as explained above in the
discussion of the "Addhost" form.  Changing it will change the next wait time
between sending ICMP Echo requests.  For example, if the current ping interval
is 5 minutes and this host has just been "pinged", if you change the ping
interval here to 5 seconds, the 5 minute wait will still have to occur before
the next "ping" of this host.  After that has occured, the next wait time
will be the new value of 5 seconds.

.TP 8
.B Community
\- This is the SNMP community name as explained above in the
discussion of the "Addhost" form.  Changing it will change the community name
which will be used for all future SNMP type queries.  If the program has
already timed out on this host and determined that SNMP is not recognized here,
you will need to "kick" it by running any SNMP function and having it
correctly respond.  Once this has been done, the program will know that this
host understands SNMP and will continue to retrieve all other SNMP information
automatically.

.TP 8
.B Last Pong
\- This represents the amount of time which has passed since the
program last received an ICMP Echo Reply from this host.  If the host is
currently down, this field should give a good indication of how long it has
been that way.

.PP
.B Netinfo
The "netinfo" form is used by the function "Netinfo()" in order to both
display network information and allow the user to make changes to it.
The fields are:

.TP 8
.B Network name
\- This is an arbitrary name which can be used to recognize the
network.  It is not used by the program in any way other than to display it.

.TP 8
.B Network number
\- This is the IP network number for this network.  It can be
determined by the logical "and" of the subnet mask and the IP address of any
host on this network.  It is displayed for convenience and can not been changed
because to do so would be to refer to a different network!

.TP 8
.B Network speed
\- This is the speed of the network.  Program tries to determine
this value by using SNMP queries of all hosts on the network as to the speed
of all interfaces connected to the network.  All of these speeds will usually
be the same value but since they may not be under some strange conditions, the
program will simply set the network speed as the highest value obtained for
it.  You may override this calculated value by typing in a new one here.
If none of the hosts on a network respond to SNMP, you will have to type in
the speed of the network here in order for the network line to be colored
(or sized on monochrome displays) correctly.

.TP 8
.B Ring radius
\- If this network is displayed as a "circle", this value is the
radius of that circle.  If the network is displayed as a "line", this value
will be ignored.

.TP 8
.B Network type
\- All networks will be displayed as a line unless you specify
otherwise.  If the network in question is a token-ring, you may want to
change this to "circle" so it will be displayed as a ring instead.
(See "Bugs" below for a discussion of the possibility of displaying "ring"
networks as an "ellipse" instead.)

.TP 8
.B Ping location of net line
\- Pressing this will toggle on and off the "pinned" flag for this network.
If this network is "pinned", then the location for the line or circle will be
specified by the user and thus not automatically calculated by the program.
If a vertical "line" network is "pinned", then it can be moved left and right
to any position you like.  If a horizontal network is "pinned", it can be
moved up and down to any position.  If a "circle" network is pinned, the
center of the circle can be placed anywhere you desire.  Specifying the network
location is done using the "Netlocation" function which is available from
the Network menu described above.

.SH OPTIONS
.TP 8
.B \-display \fIdisplay\fP
This option specifies the X server to use for the display.
.TP 8
.B \-D\fIx\fP
This option specifies what part of the program you want to display debugging
information from.  The possibilities are: m, p, f, d, h, n, i, s, x, i each
stands for the file which begins with that letter.  For example, "-Ds" would
only display debugging information from the routines contained in the source
file "snmp.c".
.bp
.SH .xsnmprc File Syntax
Here is the BNF grammar for the .xsnmprc file:

Start symbol is "xsnmp"

.nf
<xsnmp>			::= <statements>

<statements>		::= <statement> | <statements> <statement>

<statement>		::= <hostintro> <hoststmts>
			| <interfaceintro> <interfacestmts>
			| <netintro> <netstmts>
			| "include" <string>

<hostintro>		::= "host" <string>

<interfaceintro>	::= "interface" <string>

<netintro>		::= "net" <ipaddress>

<hoststmts>		::= <hoststmt> | <hoststmts> <hoststmt>

<interfacestmts>	::= <interfacestmt>
			| <interfacestmts> <interfacestmt>

<netstmts>		::= <netstmt> | <netstmts> <netstmt>

<hoststmt>		::= <mask>
			| "interval" <integer>
			| "community" <string>
			| "loc" <integer> <integer>
			| "display" <string>

<interfacestmt>		::= <mask>
			| "ipaddr" <ipaddress>

<netstmt>		::= "type" <typename>
			| "name" <string>
			| "startloc" <integer> <integer>
			| "endloc" <integer> <integer>
			| "radius" <integer>
			| "center" <integer> <integer>
			| "speed" <integer>

<typename>		::= "line" | "circle"

<mask>			::= "mask" <ipaddress>

<ipaddress> is an Internet style IP address in either
		decimal or hex notation

<string> is a quoted character string using the
		double-quote (") character.

<integer> is a decimal number.

.fi
Newline characters are skipped over the same as spaces, tabs, and commas.

Here are some examples:

.nf
host "popeye" interval 30 community "monitor" loc 100 200
	display "My system"
.fi

Specifies a host which can be looked up with the name "popeye" in order to
find the actual IP address.  This system should be "pinged" every 30 seconds.
It is in the SNMP community "monitor".  It should be positioned at
location (100, 200) ("x" and "y" coordinates respectively) on the display
window, and the name used on the display for this system should
be "My system".

.nf
net 128.111.105.0 name "ncgia network" type circle radius 20
.fi

Specifies a network which should be named "ncgia network", and should be
displayed as a circle with radius 20.  Since the exact location hasn't been
specified, it will be calculated by the program at run time.

.nf
net 128.111.235.0 name "local rt net" type line speed 10000000
.fi

Specifies a network which should be named "local rt net" and should be
displayed as a line.  Since there are no systems on this network which support
SNMP, we must specify the speed of the network if we want the line to be
displayed with the correct color or line type, thus we specify a speed
of 10Mbs.

.nf
interface "popeye" ipaddr 128.111.107.1
.fi

Specifies an additional interface for a system called "popeye".  The IP
address given is that of the additional interface.  Since the subnet mask
is the same as the builtin default, we need not specify it.  This should only
be necessary for systems which are gateways between subnets and which do
not support SNMP, and thus can not be queried for their additional interfaces.

.nf
include "othernets.conf"
.fi

Specifies an additional configuration file from which more information can
be read.

.SH "SEE ALSO"
X(1)
.SH BUGS
The program should allow networks to be displayed as an ellipse as well as
a circle and a line.  Functionality for specifying the major and minor axes
wouldn't be hard to implement.  The hard part is calculating the X and Y
location where a line drawn from the host location in the direction of the
center of the ellipse would meet the ellipse.  To do this you need to find the
solution to the two simultaneous equations, one for the line and one for the
ellipse.  The equation for the line is, of course,
.sp
.ce
.EQ
.ds 11 "\f2y\fP
.ds 12 "\f2\(eq\fP
.as 11 "\^\*(12
.ds 12 "\f2mx\fP
.as 11 "\*(12
.ds 12 "\f2\(pl\fP
.as 11 "\^\*(12
.ds 12 "\f2b\fP
.as 11 "\*(12
.if 1m>\n(.v .ne 1m
.rn 11 10
\&\*(10
.EN
.lf 629
.sp
where m is
the slope of the line and b is the y-intercept.  The values of m and b
are trivial to calculate.  The equation for the ellipse
is, of course
.sp
.ce
.EQ
.ds 11 "\f2x\fP
.ds 12 "\f2a\fP
.nr 11 0\w'\s+0\*(11'
.nr 12 0\w'\s+0\*(12'
.nr 13 \n(11
.if \n(12>\n(13 .nr 13 \n(12
.nr 13 \n(13+.5m
.ds 11 \v'0.7m'\h'\n(13u-\n(12u/2u'\*(12\
\h'-\n(12u-\n(11u/2u'\v'-1.3m'\*(11\
\h'-\n(13u-\n(11u/2u+.1m'\v'0.3m'\l'\n(13u-.2m'\h'.1m'\v'0.3m'
.ds 11 \|\b'\(lt\(bv\(lb'\*(11\b'\(rt\(bv\(rb'
.ds 12 "\f12\fP
.as 11 \v'-1.34m'\s-3\*(12\s+3\^\v'1.34m'
.ds 12 "\f2*\fP
.as 11 "\*(12
.ds 12 "\f2y\fP
.ds 13 "\f2b\fP
.nr 12 0\w'\s+0\*(12'
.nr 13 0\w'\s+0\*(13'
.nr 14 \n(12
.if \n(13>\n(14 .nr 14 \n(13
.nr 14 \n(14+.5m
.ds 12 \v'0.7m'\h'\n(14u-\n(13u/2u'\*(13\
\h'-\n(13u-\n(12u/2u'\v'-1.3m'\*(12\
\h'-\n(14u-\n(12u/2u+.1m'\v'0.3m'\l'\n(14u-.2m'\h'.1m'\v'0.3m'
.ds 12 \|\b'\(lt\(bv\(lb'\*(12\b'\(rt\(bv\(rb'
.ds 13 "\f12\fP
.as 12 \v'-1.34m'\s-3\*(13\s+3\^\v'1.34m'
.as 11 "\*(12
.ds 12 "\f2\(eq\fP
.as 11 "\*(12
.ds 12 "\f11\fP
.as 11 "\*(12
.ds 11 \x'0-0.7m'\*(11\x'0.9m'
.if 3m>\n(.v .ne 3m
.rn 11 10
\&\*(10
.EN
.lf 639
.sp
where a is the size of the axis aligned in the X direction and b is the
size of the axis aligned in the Y direction.  All you need do is solve both
of these equations for x and y.  In both cases, the answers should involve
a square root since there are two possible solutions for each.  Just send me
the solutions and I'll be glad to program up the rest!  Good luck!
.PP
Lines must be either horizontal or vertical only.  I didn't want to deal with
calculating the closest point on a line to a specified point.  If you want to
write this code and send it to me, I'll be glad to incorporate it and
acknowledge your contribution.
.SH AUTHOR
Richard Johnson
.br
UCSB National Center for Geographic Information and Analysis
.br
Santa Barbara, Ca.
.br
<raj@ncgia.ucsb.edu>
.br
(Along with the contributions of others as acknowledged in the CHANGES
file of the distribution.)
