%{
#include "xsnmp.h"

#define Debug	(Flags & (PROG_DEBUG_YACC | PROG_DEBUG))

extern int	yylineno;

PARSEHOST	*_host;
PARSENET	*_net;
PARSEIF		*_if;

HOSTENTRY	*_hptr=(HOSTENTRY *)NULL;
NETENTRY	*_nptr=(NETENTRY *)NULL;
IFENTRY		*_ifptr=(IFENTRY *)NULL;

%}

%union {
	int	number;
	char	*string;
};

%token <number> HOST
%token <number> MASK
%token <number> INTERVAL
%token <number> COMMUNITY
%token <number> LOC
%token <number>	SHOWNAME
%token <number>	DISPLAY

%token <number>	INTERFACE
%token <number>	IPADDR

%token <number>	NET
%token <number>	NAME
%token <number>	TYPE
%token <number>	CIRCLE
%token <number>	LINE
%token <number>	STARTLOC
%token <number>	ENDLOC
%token <number>	RADIUS
%token <number>	CENTER
%token <number>	SPEED
%token <number>	HIDE
%token <number>	SHOW

%token <number>	INCLUDE

%token <string>	STRING
%token <number>	HEXNUMBER
%token <number> INTEGER
%token <number>	IPADDRESS

%type <number>	ipaddress
%type <number>	mask
%type <number>	typename

%start xsnmp

%%

xsnmp		:	statements
		;

statements	:	statement
		|	statements statement


statement	:	hostintro
			hoststmts	{
			    _hptr = addhost(_host->pah_name, _host->pah_mask
				    , _host->pah_community
				    , _host->pah_interval, _host->pah_display);
			    if(_hptr) {
				movehost(_hptr, _host->pah_x, _host->pah_y);
				_hptr->hl_flags |= _host->pah_flags;
			    }
			    freeparsehost(_host);
			    _host = (PARSEHOST *)NULL;
			}
		|	interfaceintro
    			interfacestmts	{
			    if(Debug)
				printf("interface\n");
			    if((_hptr = nametohost(_if->pai_name))
			       == (HOSTENTRY *)NULL)
				yyerror("no such host");
			    else {
				_ifptr = addif(_hptr, _if->pai_addr
					       , _if->pai_mask);
				if(_ifptr)
				    _ifptr->if_flags |= _if->pai_flags;
			    }
			    freeparseif(_if);
			    _if = (PARSEIF *)NULL;
			}
		|	netintro
			netstmts	{
			    if(Debug)
				printf("net '%s'\n", _net->pan_name);
			    if(iftonet(_net->pan_addr) != (NETENTRY *)NULL)
				yyerror("2nd occurance of net ignored");
			    else {
				if(Debug)
				    printf("addr %s\n"
					   , inet_ntoa(_net->pan_addr));
				_nptr = newnetentry(_net->pan_addr
						    , _net->pan_speed);
				if(_nptr) {
				    if(_net->pan_name)
					setnetname(_nptr, _net->pan_name);
				    if(_net->pan_startx > 0
				       && _net->pan_starty > 0
				       && _net->pan_endx > 0
				       && _net->pan_endy > 0) {
					if(Debug)
					    printf("net line\n");
					setnetline(_nptr, _net->pan_startx
						  , _net->pan_starty
						  , _net->pan_endx
						  , _net->pan_endy);
					_nptr->nl_flags |= NET_PINNED;
				    } else if(_net->pan_centerx > 0
					      || _net->pan_centery > 0
					      || _net->pan_radius > 0) {
					if(Debug)
					    printf("net circle\n");
					setnetcircle(_nptr, _net->pan_centerx
						     , _net->pan_centery
						     , _net->pan_radius);
					if(_net->pan_centerx > 0
					   && _net->pan_centery > 0)
					    _nptr->nl_flags |= NET_PINNED;
				    }
				    _nptr->nl_flags |= _net->pan_flags;
				}
			    }
			    freeparsenet(_net);
			    _net = (PARSENET *)NULL;
			}
		|	INCLUDE
			STRING		{
			    if(Debug)
				printf("Include '%s'\n", $2);
			    doinclude($2);
			}
		;

hostintro	:	HOST
			STRING		{
			    if(Debug)
				printf("Host '%s'\n", $2);
			    _host = newparsehost($2, in_include()
						 ? FL_INCLUDE : 0);
			}
		;

interfaceintro	:	INTERFACE
    			STRING		{
			    if(Debug)
				printf("Additional interface for '%s'\n", $2);
			    _if = newparseif($2, in_include()
					     ? FL_INCLUDE : 0);
			}

netintro	:	NET
			IPADDRESS	{
			    if(Debug)
				printf("Network Ipaddress 0x%x\n", $2);
			    _net = newparsenet($2, in_include()
					       ? NET_INCLUDE : 0);
			}
		;

hoststmts	:	hoststmt
		|	hoststmts hoststmt
		;

interfacestmts	:	interfacestmt
		|	interfacestmts interfacestmt
		;

netstmts	:	netstmt
		|	netstmts netstmt
		;

hoststmt	:	mask		{
			    if(Debug)
				printf("Host mask: 0x%x\n", $1);
			    if(! _host)
				yyerror("mask with no host");
			    else {
				_host->pah_mask = $1;
				_host->pah_flags |= FL_CONFMASK;
			    }
			}

		|	INTERVAL
			INTEGER		{
			    if(Debug)
				printf("Interval: %d\n", $2);
			    if(! _host)
				yyerror("interval with no host");
			    else
				_host->pah_interval = $2;
			}
		|	COMMUNITY
			STRING		{
			    if(Debug)
				printf("Community '%s'\n", $2);
			    if(! _host)
				yyerror("community with no host");
			    else {
				if(_host->pah_community)
				    free(_host->pah_community);
				_host->pah_community = cpystr($2);
			    }
			}
		|	LOC
			INTEGER
			INTEGER		{
			    if(Debug)
				printf("Loc %d %d\n", $2, $3);
			    if(! _host)
				yyerror("loc with no host");
			    else {
				_host->pah_x = $2;
				_host->pah_y = $3;
			    }
			}
		|	DISPLAY
    			STRING		{
			    if(Debug)
				printf("displayname '%s'\n", $2);
			    if(! _host)
				yyerror("displayname with no host");
			    else {
				if(_host->pah_display)
				    free(_host->pah_display);
				_host->pah_display = cpystr($2);
			    }
			}
		;

interfacestmt	:	mask		{
			    if(Debug)
				printf("Interface mask: 0x%x\n", $1);
			    if(! _if) yyerror("mask with no interface");
			    else {
				_if->pai_mask = $1;
				_if->pai_flags |= FL_CONFMASK;
			    }
			}
		|	IPADDR
    			IPADDRESS	{
			    if(Debug)
				printf("Ipaddress 0x%x\n", $2);
			    if(! _if)
				yyerror("ipaddr with no hostname");
			    else
				_if->pai_addr = $2;
			}
		;
		
netstmt		:	TYPE
			typename
		|	NAME
    			STRING		{
			    if(! _net)
				yyerror("net name with no ipaddr");
			    else {
				if(_net->pan_name) free(_net->pan_name);
				_net->pan_name = cpystr($2);
			    }
			}
		|	STARTLOC
			INTEGER
			INTEGER		{
			    if(Debug)
				printf("Startloc %d %d\n", $2, $3);
			    if(! _net)
				yyerror("startloc with no network");
			    else {
				_net->pan_startx = $2;
				_net->pan_starty = $3;
			    }
			}
		|	ENDLOC
			INTEGER
			INTEGER		{
			    if(Debug)
				printf("Endloc %d %d\n", $2, $3);
			    if(! _net)
				yyerror("endloc with no network");
			    else {
				_net->pan_endx = $2;
				_net->pan_endy = $3;
			    }
			}
		|	RADIUS
			INTEGER		{
			    if(Debug)
				printf("Radius %d\n", $2);
			    if(! _net)
				yyerror("radius with no network");
			    else
				_net->pan_radius = $2;
			}
		|	CENTER
			INTEGER
			INTEGER		{
			    if(Debug)
				printf("Center %d %d\n", $2, $3);
			    if(! _net) yyerror("center with no network");
			    _net->pan_centerx = $2;
			    _net->pan_centery = $3;
			}
		|	SPEED
			INTEGER		{
			    if(Debug)
				printf("Speed %d\n", $2);
			    if(! _net)
				yyerror("speed with no network");
			    else {
				_net->pan_speed = $2;
				_net->pan_flags |= NET_SPEEDSET;
			    }
			}
		;

typename	:	LINE		{
    			    $$ = $1;
			    if(Debug)
				printf("Network should be a line\n");
			}
    		|	CIRCLE		{
		    	    $$ = $1;
			    if(! _net)
				yyerror("type with no network");
			    else
				_net->pan_flags |= NET_CIRCLE;
			}
		;

mask		:	MASK
			ipaddress	{
			    if(Debug)
				printf("Mask: 0x%x\n", $2);
			    $$ = $2;
			}

ipaddress	:	IPADDRESS	{
			    if(Debug)
				printf("Ipaddr: 0x%x\n", $1);
			    $$ = $1;
			}				    
		|	HEXNUMBER	{
			    if(Debug)
				printf("Hexnumber: 0x%x\n", $1);
			    $$ = $1;
			}
		;					
%%

yyerror(string) char *string; {

    printf("Got an error on or before line %d:\n%s\n", yylineno, string);
}

yywrap() {

    if(Debug)
	printf("yywrap called\n");
    return(1);		/* success */
}

#include "lex.yy.c"
